@@ -21,12 +21,6 @@ while getopts ":b:v:" opt; do
 	esac
 done
 
-if [ ! -z "${VERSION}" ]; then
-  ln -frs ${BUILDDIR}/pipewire-pulseaudio/src/libpulse-pw.so.${VERSION} ${BUILDDIR}/pipewire-pulseaudio/src/libpulse.so.0
-  ln -frs ${BUILDDIR}/pipewire-pulseaudio/src/libpulse-mainloop-glib-pw.so.${VERSION} ${BUILDDIR}/pipewire-pulseaudio/src/libpulse-mainloop-glib.so.0
-  ln -frs ${BUILDDIR}/pipewire-jack/src/libjack-pw.so.${VERSION} ${BUILDDIR}/pipewire-jack/src/libjack.so.0
-fi
-
 if [ -z "${BUILDDIR}" ]; then
 	BUILDDIR=${PWD}/build
 	echo "Using default build directory: ${BUILDDIR}"
@@ -37,12 +31,14 @@ if [ ! -d ${BUILDDIR} ]; then
 	exit -1
 fi
 
-export PIPEWIRE_CONFIG_FILE="${BUILDDIR}/src/daemon/pipewire.conf"
+export PIPEWIRE_CONFIG_FILE="${BUILDDIR}/src/daemon/pipewire-uninstalled.conf"
 export SPA_PLUGIN_DIR="${BUILDDIR}/spa/plugins"
 export PIPEWIRE_MODULE_DIR="${BUILDDIR}/src/modules"
 export PATH="${BUILDDIR}/src/daemon:${BUILDDIR}/src/tools:${BUILDDIR}/src/examples:${PATH}"
-export LD_LIBRARY_PATH="${BUILDDIR}/pipewire-pulseaudio/src/:${BUILDDIR}/src/pipewire/:${BUILDDIR}/pipewire-jack/src/:${LD_LIBRARY_PATH}"
-export GST_PLUGIN_PATH="${BUILDDIR}/src/gst/:${GST_PLUGIN_PATH}"
+export LD_LIBRARY_PATH="${BUILDDIR}/pipewire-pulseaudio/src/:${BUILDDIR}/src/pipewire/:${BUILDDIR}/pipewire-jack/src/${LD_LIBRARY_PATH+":$LD_LIBRARY_PATH"}"
+export GST_PLUGIN_PATH="${BUILDDIR}/src/gst/${GST_PLUGIN_PATH+":${GST_PLUGIN_PATH}"}"
+export ACP_PATHS_DIR=${PWD}"/spa/plugins/alsa/mixer/paths"
+export ACP_PROFILES_DIR=${PWD}"/spa/plugins/alsa/mixer/profile-sets"
 
 # FIXME: find a nice, shell-neutral way to specify a prompt
 ${SHELL}
