@@ -26,7 +26,9 @@
 #include <stdio.h>
 #include <errno.h>
 #include <math.h>
+#include <getopt.h>
 #include <time.h>
+#include <signal.h>
 
 #include "config.h"
 
@@ -40,7 +42,9 @@
 #include <spa/monitor/device.h>
 
 #include "pipewire/pipewire.h"
+#include "pipewire/private.h"
 #include "extensions/session-manager.h"
+#include "extensions/client-node.h"
 
 #include <dbus/dbus.h>
 
@@ -52,6 +56,7 @@
 
 #define sm_object_emit_update(s)		sm_object_emit(s, update, 0)
 #define sm_object_emit_destroy(s)		sm_object_emit(s, destroy, 0)
+#define sm_object_emit_free(s)			sm_object_emit(s, free, 0)
 
 #define sm_media_session_emit(s,m,v,...) spa_hook_list_call(&(s)->hooks, struct sm_media_session_events, m, v, ##__VA_ARGS__)
 
@@ -61,9 +66,12 @@
 #define sm_media_session_emit_rescan(s,seq)		sm_media_session_emit(s, rescan, 0, seq)
 #define sm_media_session_emit_destroy(s)		sm_media_session_emit(s, destroy, 0)
 
+int sm_access_flatpak_start(struct sm_media_session *sess);
+int sm_access_portal_start(struct sm_media_session *sess);
 int sm_metadata_start(struct sm_media_session *sess);
 int sm_alsa_midi_start(struct sm_media_session *sess);
 int sm_v4l2_monitor_start(struct sm_media_session *sess);
+int sm_libcamera_monitor_start(struct sm_media_session *sess);
 int sm_bluez5_monitor_start(struct sm_media_session *sess);
 int sm_alsa_monitor_start(struct sm_media_session *sess);
 int sm_suspend_node_start(struct sm_media_session *sess);
@@ -115,6 +123,8 @@ struct impl {
 	struct spa_list endpoint_link_list;	/** list of struct endpoint_link */
 	struct pw_map endpoint_links;		/** map of endpoint_link */
 
+	struct spa_list link_list;		/** list of struct link */
+
 	struct spa_list sync_list;		/** list of struct sync */
 	int rescan_seq;
 	int last_seq;
@@ -141,7 +151,8 @@ struct link {
 	uint32_t input_port;
 
 	struct endpoint_link *endpoint_link;
-	struct spa_list link;		/**< link in struct endpoint_link link_list */
+	struct spa_list link;		/**< link in struct endpoint_link link_list or
+					  *  struct impl link_list */
 };
 
 struct object_info {
@@ -174,12 +185,14 @@ static void remove_object(struct impl *impl, struct sm_object *obj)
 	obj->id = SPA_ID_INVALID;
 }
 
-static void *find_object(struct impl *impl, uint32_t id)
+static void *find_object(struct impl *impl, uint32_t id, const char *type)
 {
-	void *obj;
-	if ((obj = pw_map_lookup(&impl->globals, id)) != NULL)
-		return obj;
-	return NULL;
+	struct sm_object *obj;
+	if ((obj = pw_map_lookup(&impl->globals, id)) == NULL)
+		return NULL;
+	if (type != NULL && strcmp(obj->type, type) != 0)
+		return NULL;
+	return obj;
 }
 
 static struct data *object_find_data(struct sm_object *obj, const char *id)
@@ -234,17 +247,56 @@ int sm_object_remove_data(struct sm_object *obj, const char *id)
 
 int sm_object_destroy(struct sm_object *obj)
 {
-	pw_log_debug(NAME" %p: object %d", obj->session, obj->id);
-	if (obj->proxy) {
-		pw_proxy_destroy(obj->proxy);
-		if (obj->handle == obj->proxy)
-			obj->handle = NULL;
-		obj->proxy = NULL;
+	struct impl *impl = SPA_CONTAINER_OF(obj->session, struct impl, this);
+	struct data *d;
+	struct pw_proxy *p, *h;
+
+	p = obj->proxy;
+	h = obj->handle;
+
+	pw_log_debug(NAME" %p: object %d proxy:%p handle:%p", obj->session,
+			obj->id, p, h);
+
+	sm_object_emit_destroy(obj);
+
+	if (SPA_FLAG_IS_SET(obj->mask, SM_OBJECT_CHANGE_MASK_LISTENER)) {
+		SPA_FLAG_CLEAR(obj->mask, SM_OBJECT_CHANGE_MASK_LISTENER);
+		spa_hook_remove(&obj->object_listener);
 	}
-	if (obj->handle) {
-		pw_proxy_destroy(obj->handle);
-		obj->handle = NULL;
+
+	if (obj->id != SPA_ID_INVALID)
+		remove_object(impl, obj);
+
+	if (obj->destroy)
+		obj->destroy(obj);
+
+	if (p) {
+		pw_proxy_ref(p);
+		spa_hook_remove(&obj->proxy_listener);
 	}
+	if (h) {
+		pw_proxy_ref(h);
+		spa_hook_remove(&obj->handle_listener);
+	}
+	if (p)
+		pw_proxy_destroy(p);
+	if (h != p)
+		pw_proxy_destroy(h);
+
+	sm_object_emit_free(obj);
+
+	if (obj->props)
+		pw_properties_free(obj->props);
+	obj->props = NULL;
+
+	spa_list_consume(d, &obj->data, link) {
+		spa_list_remove(&d->link);
+		free(d);
+	}
+	if (p)
+		pw_proxy_unref(p);
+	if (h)
+		pw_proxy_unref(h);
 	return 0;
 }
 
@@ -289,6 +341,36 @@ static uint32_t clear_params(struct spa_list *param_list, uint32_t id)
 	return count;
 }
 
+/**
+ * Core
+ */
+static const struct object_info core_object_info = {
+	.type = PW_TYPE_INTERFACE_Core,
+	.version = PW_VERSION_CORE,
+	.size = sizeof(struct sm_object),
+	.init = NULL,
+};
+
+/**
+ * Module
+ */
+static const struct object_info module_info = {
+	.type = PW_TYPE_INTERFACE_Module,
+	.version = PW_VERSION_MODULE,
+	.size = sizeof(struct sm_object),
+	.init = NULL,
+};
+
+/**
+ * Factory
+ */
+static const struct object_info factory_info = {
+	.type = PW_TYPE_INTERFACE_Factory,
+	.version = PW_VERSION_FACTORY,
+	.size = sizeof(struct sm_object),
+	.init = NULL,
+};
+
 /**
  * Clients
  */
@@ -437,7 +519,8 @@ static void node_event_info(void *object, const struct pw_node_info *info)
 			case SPA_PARAM_PropInfo:
 			case SPA_PARAM_Props:
 			case SPA_PARAM_EnumFormat:
-				subscribe[n_subscribe++] = info->params[i].id;
+				if (info->params[i].flags & SPA_PARAM_INFO_READ)
+					subscribe[n_subscribe++] = info->params[i].id;
 				break;
 			default:
 				break;
@@ -495,7 +578,7 @@ static int node_init(void *object)
 
 	if (props) {
 		if ((str = pw_properties_get(props, PW_KEY_DEVICE_ID)) != NULL)
-			node->device = find_object(impl, atoi(str));
+			node->device = find_object(impl, atoi(str), NULL);
 		pw_log_debug(NAME" %p: node %d parent device %s (%p)", impl,
 				node->obj.id, str, node->device);
 		if (node->device) {
@@ -559,6 +642,16 @@ static const struct pw_port_events port_events = {
 	.info = port_event_info,
 };
 
+static enum spa_audio_channel find_channel(const char *name)
+{
+        int i;
+        for (i = 0; spa_type_audio_channel[i].name; i++) {
+                if (strcmp(name, spa_debug_type_short_name(spa_type_audio_channel[i].name)) == 0)
+                        return spa_type_audio_channel[i].type;
+        }
+        return SPA_AUDIO_CHANNEL_UNKNOWN;
+}
+
 static int port_init(void *object)
 {
 	struct sm_port *port = object;
@@ -570,11 +663,19 @@ static int port_init(void *object)
 		if ((str = pw_properties_get(props, PW_KEY_PORT_DIRECTION)) != NULL)
 			port->direction = strcmp(str, "out") == 0 ?
 				PW_DIRECTION_OUTPUT : PW_DIRECTION_INPUT;
+		if ((str = pw_properties_get(props, PW_KEY_FORMAT_DSP)) != NULL) {
+			if (strcmp(str, "32 bit float mono audio") == 0)
+				port->type = SM_PORT_TYPE_DSP_AUDIO;
+			else if (strcmp(str, "8 bit raw midi") == 0)
+				port->type = SM_PORT_TYPE_DSP_MIDI;
+		}
+		if ((str = pw_properties_get(props, PW_KEY_AUDIO_CHANNEL)) != NULL)
+			port->channel = find_channel(str);
 		if ((str = pw_properties_get(props, PW_KEY_NODE_ID)) != NULL)
-			port->node = find_object(impl, atoi(str));
+			port->node = find_object(impl, atoi(str), PW_TYPE_INTERFACE_Node);
 
-		pw_log_debug(NAME" %p: port %d parent node %s (%p) direction:%d", impl,
-				port->obj.id, str, port->node, port->direction);
+		pw_log_debug(NAME" %p: port %d parent node %s (%p) direction:%d type:%d", impl,
+				port->obj.id, str, port->node, port->direction, port->type);
 		if (port->node) {
 			spa_list_append(&port->node->port_list, &port->link);
 			port->node->obj.avail |= SM_NODE_CHANGE_MASK_PORTS;
@@ -725,7 +826,7 @@ static int endpoint_init(void *object)
 
 	if (props) {
 		if ((str = pw_properties_get(props, PW_KEY_SESSION_ID)) != NULL)
-			endpoint->session = find_object(impl, atoi(str));
+			endpoint->session = find_object(impl, atoi(str), PW_TYPE_INTERFACE_Session);
 		pw_log_debug(NAME" %p: endpoint %d parent session %s", impl,
 				endpoint->obj.id, str);
 		if (endpoint->session) {
@@ -809,7 +910,7 @@ static int endpoint_stream_init(void *object)
 
 	if (props) {
 		if ((str = pw_properties_get(props, PW_KEY_ENDPOINT_ID)) != NULL)
-			stream->endpoint = find_object(impl, atoi(str));
+			stream->endpoint = find_object(impl, atoi(str), PW_TYPE_INTERFACE_Endpoint);
 		pw_log_debug(NAME" %p: stream %d parent endpoint %s", impl,
 				stream->obj.id, str);
 		if (stream->endpoint) {
@@ -907,37 +1008,6 @@ static const struct object_info endpoint_link_info = {
 /**
  * Proxy
  */
-static void
-destroy_proxy(void *data)
-{
-	struct sm_object *obj = data;
-	struct impl *impl = SPA_CONTAINER_OF(obj->session, struct impl, this);
-	struct data *d;
-
-	pw_log_debug("object %p: proxy:%p id:%d", obj, obj->proxy, obj->id);
-
-	spa_hook_remove(&obj->proxy_listener);
-	if (SPA_FLAG_IS_SET(obj->mask, SM_OBJECT_CHANGE_MASK_LISTENER))
-		spa_hook_remove(&obj->object_listener);
-
-	if (obj->id != SPA_ID_INVALID)
-		remove_object(impl, obj);
-
-	sm_object_emit_destroy(obj);
-
-	if (obj->destroy)
-		obj->destroy(obj);
-
-	if (obj->props)
-		pw_properties_free(obj->props);
-	obj->props = NULL;
-
-	spa_list_consume(d, &obj->data, link) {
-		spa_list_remove(&d->link);
-		free(d);
-	}
-}
-
 static void done_proxy(void *data, int seq)
 {
 	struct sm_object *obj = data;
@@ -967,7 +1037,6 @@ static void bound_proxy(void *data, uint32_t id)
 
 static const struct pw_proxy_events proxy_events = {
 	PW_VERSION_PROXY_EVENTS,
-	.destroy = destroy_proxy,
 	.done = done_proxy,
 	.bound = bound_proxy,
 };
@@ -983,7 +1052,13 @@ static const struct object_info *get_object_info(struct impl *impl, const char *
 {
 	const struct object_info *info;
 
-	if (strcmp(type, PW_TYPE_INTERFACE_Client) == 0)
+	if (strcmp(type, PW_TYPE_INTERFACE_Core) == 0)
+		info = &core_object_info;
+	else if (strcmp(type, PW_TYPE_INTERFACE_Module) == 0)
+		info = &module_info;
+	else if (strcmp(type, PW_TYPE_INTERFACE_Factory) == 0)
+		info = &factory_info;
+	else if (strcmp(type, PW_TYPE_INTERFACE_Client) == 0)
 		info = &client_info;
 	else if (strcmp(type, SPA_TYPE_INTERFACE_Device) == 0)
 		info = &spa_device_info;
@@ -1032,9 +1107,7 @@ static struct sm_object *init_object(struct impl *impl, const struct object_info
 			pw_proxy_add_object_listener(obj->proxy, &obj->object_listener, info->events, obj);
 		SPA_FLAG_UPDATE(obj->mask, SM_OBJECT_CHANGE_MASK_LISTENER, info->events != NULL);
 	}
-	if (handle) {
-		pw_proxy_add_listener(obj->handle, &obj->handle_listener, &proxy_events, obj);
-	}
+	pw_proxy_add_listener(obj->handle, &obj->handle_listener, &proxy_events, obj);
 
 	if (info->init)
 		info->init(obj);
@@ -1055,6 +1128,9 @@ create_object(struct impl *impl, struct pw_proxy *proxy, struct pw_proxy *handle
 
 	type = pw_proxy_get_type(handle, NULL);
 
+	if (strcmp(type, PW_TYPE_INTERFACE_ClientNode) == 0)
+		type = PW_TYPE_INTERFACE_Node;
+
 	info = get_object_info(impl, type);
 	if (info == NULL) {
 		pw_log_error(NAME" %p: unknown object type %s", impl, type);
@@ -1063,7 +1139,8 @@ create_object(struct impl *impl, struct pw_proxy *proxy, struct pw_proxy *handle
 	}
 	obj = init_object(impl, info, proxy, handle, SPA_ID_INVALID, props);
 
-	pw_log_debug(NAME" %p: created new object %p proxy %p", impl, obj, obj->proxy);
+	pw_log_debug(NAME" %p: created new object %p proxy:%p handle:%p", impl,
+			obj, obj->proxy, obj->handle);
 
 	return obj;
 }
@@ -1141,7 +1218,7 @@ registry_global(void *data, uint32_t id,
 	if (info == NULL)
 		return;
 
-	obj = find_object(impl, id);
+	obj = find_object(impl, id, NULL);
 	if (obj == NULL) {
 		bind_object(impl, info, id, permissions, type, version, props);
 	} else {
@@ -1178,7 +1255,29 @@ int sm_media_session_add_listener(struct sm_media_session *sess, struct spa_hook
 struct sm_object *sm_media_session_find_object(struct sm_media_session *sess, uint32_t id)
 {
 	struct impl *impl = SPA_CONTAINER_OF(sess, struct impl, this);
-	return find_object(impl, id);
+	return find_object(impl, id, NULL);
+}
+
+int sm_media_session_destroy_object(struct sm_media_session *sess, uint32_t id)
+{
+	struct impl *impl = SPA_CONTAINER_OF(sess, struct impl, this);
+	pw_registry_destroy(impl->registry, id);
+	return 0;
+}
+
+int sm_media_session_for_each_object(struct sm_media_session *sess,
+                            int (*callback) (void *data, struct sm_object *object),
+                            void *data)
+{
+	struct impl *impl = SPA_CONTAINER_OF(sess, struct impl, this);
+	struct sm_object *obj;
+	int res;
+
+	spa_list_for_each(obj, &impl->global_list, link) {
+		if ((res = callback(data, obj)) != 0)
+			return res;
+	}
+	return 0;
 }
 
 int sm_media_session_schedule_rescan(struct sm_media_session *sess)
@@ -1250,10 +1349,10 @@ registry_global_remove(void *data, uint32_t id)
 
 	pw_log_debug(NAME " %p: remove global '%d'", impl, id);
 
-	if ((obj = find_object(impl, id)) == NULL)
+	if ((obj = find_object(impl, id, NULL)) == NULL)
 		return;
 
-	remove_object(impl, obj);
+	sm_object_destroy(obj);
 }
 
 static const struct pw_registry_events registry_events = {
@@ -1373,12 +1472,19 @@ static void check_endpoint_link(struct endpoint_link *link)
 	}
 }
 
+static void proxy_link_removed(void *data)
+{
+	struct link *l = data;
+	pw_proxy_destroy(l->proxy);
+}
+
 static void proxy_link_destroy(void *data)
 {
 	struct link *l = data;
 
+	spa_list_remove(&l->link);
+
 	if (l->endpoint_link) {
-		spa_list_remove(&l->link);
 		check_endpoint_link(l->endpoint_link);
 		l->endpoint_link = NULL;
 	}
@@ -1386,9 +1492,59 @@ static void proxy_link_destroy(void *data)
 
 static const struct pw_proxy_events proxy_link_events = {
 	PW_VERSION_PROXY_EVENTS,
+	.removed = proxy_link_removed,
 	.destroy = proxy_link_destroy
 };
 
+static int score_ports(struct sm_port *out, struct sm_port *in)
+{
+	int score = 0;
+
+	if (in->direction != PW_DIRECTION_INPUT || out->direction != PW_DIRECTION_OUTPUT)
+		return 0;
+
+	if (out->type != SM_PORT_TYPE_UNKNOWN && in->type != SM_PORT_TYPE_UNKNOWN &&
+	    in->type != out->type)
+		return 0;
+
+	if (out->channel == in->channel)
+		score += 100;
+	else if ((out->channel == SPA_AUDIO_CHANNEL_SL && in->channel == SPA_AUDIO_CHANNEL_RL) ||
+	         (out->channel == SPA_AUDIO_CHANNEL_RL && in->channel == SPA_AUDIO_CHANNEL_SL) ||
+	         (out->channel == SPA_AUDIO_CHANNEL_SR && in->channel == SPA_AUDIO_CHANNEL_RR) ||
+	         (out->channel == SPA_AUDIO_CHANNEL_RR && in->channel == SPA_AUDIO_CHANNEL_SR))
+		score += 60;
+	else if ((out->channel == SPA_AUDIO_CHANNEL_FC && in->channel == SPA_AUDIO_CHANNEL_MONO) ||
+	         (out->channel == SPA_AUDIO_CHANNEL_MONO && in->channel == SPA_AUDIO_CHANNEL_FC))
+		score += 50;
+	else if (in->channel == SPA_AUDIO_CHANNEL_UNKNOWN ||
+	    in->channel == SPA_AUDIO_CHANNEL_MONO ||
+	    out->channel == SPA_AUDIO_CHANNEL_UNKNOWN ||
+	    out->channel == SPA_AUDIO_CHANNEL_MONO)
+		score += 10;
+	if (score > 0 && !in->visited)
+		score += 5;
+	if (score <= 10)
+		score = 0;
+	return score;
+}
+
+static struct sm_port *find_input_port(struct impl *impl, struct sm_node *outnode,
+		struct sm_port *outport, struct sm_node *innode)
+{
+	struct sm_port *inport, *best_port = NULL;
+	int score, best_score = 0;
+
+	spa_list_for_each(inport, &innode->port_list, link) {
+		score = score_ports(outport, inport);
+		if (score > best_score) {
+			best_score = score;
+			best_port = inport;
+		}
+	}
+	return best_port;
+}
+
 static int link_nodes(struct impl *impl, struct endpoint_link *link,
 		struct sm_node *outnode, struct sm_node *innode)
 {
@@ -1401,51 +1557,52 @@ static int link_nodes(struct impl *impl, struct endpoint_link *link,
 	pw_properties_setf(props, PW_KEY_LINK_OUTPUT_NODE, "%d", outnode->obj.id);
 	pw_properties_setf(props, PW_KEY_LINK_INPUT_NODE, "%d", innode->obj.id);
 
-	for (outport = spa_list_first(&outnode->port_list, struct sm_port, link),
-	    inport = spa_list_first(&innode->port_list, struct sm_port, link);
-	    !spa_list_is_end(outport, &outnode->port_list, link) &&
-	    !spa_list_is_end(inport, &innode->port_list, link);) {
+	spa_list_for_each(inport, &innode->port_list, link)
+		inport->visited = false;
+
+	spa_list_for_each(outport, &outnode->port_list, link) {
+		struct link *l;
+		struct pw_proxy *p;
+
+		if (outport->direction != PW_DIRECTION_OUTPUT)
+			continue;
+
+		inport = find_input_port(impl, outnode, outport, innode);
+		if (inport == NULL) {
+			pw_log_debug(NAME" %p: port %d:%d can't be linked", impl,
+				outport->direction, outport->obj.id);
+			continue;
+		}
+		inport->visited = true;
 
 		pw_log_debug(NAME" %p: port %d:%d -> %d:%d", impl,
 				outport->direction, outport->obj.id,
 				inport->direction, inport->obj.id);
 
-		if (outport->direction == PW_DIRECTION_OUTPUT &&
-		    inport->direction == PW_DIRECTION_INPUT) {
-			struct link *l;
-			struct pw_proxy *p;
+		pw_properties_setf(props, PW_KEY_LINK_OUTPUT_PORT, "%d", outport->obj.id);
+		pw_properties_setf(props, PW_KEY_LINK_INPUT_PORT, "%d", inport->obj.id);
 
-			pw_properties_setf(props, PW_KEY_LINK_OUTPUT_PORT, "%d", outport->obj.id);
-			pw_properties_setf(props, PW_KEY_LINK_INPUT_PORT, "%d", inport->obj.id);
+		p = pw_core_create_object(impl->policy_core,
+					"link-factory",
+					PW_TYPE_INTERFACE_Link,
+					PW_VERSION_LINK,
+					&props->dict, sizeof(struct link));
+		if (p == NULL)
+			return -errno;
 
-			p = pw_core_create_object(impl->policy_core,
-						"link-factory",
-						PW_TYPE_INTERFACE_Link,
-						PW_VERSION_LINK,
-						&props->dict, sizeof(struct link));
-			if (p == NULL)
-				return -errno;
+		l = pw_proxy_get_user_data(p);
+		l->proxy = p;
+		l->output_node = outnode->obj.id;
+		l->output_port = outport->obj.id;
+		l->input_node = innode->obj.id;
+		l->input_port = inport->obj.id;
+		pw_proxy_add_listener(p, &l->listener, &proxy_link_events, l);
 
-			l = pw_proxy_get_user_data(p);
-			l->proxy = p;
-			l->output_node = outnode->obj.id;
-			l->output_port = outport->obj.id;
-			l->input_node = innode->obj.id;
-			l->input_port = inport->obj.id;
-			pw_proxy_add_listener(p, &l->listener, &proxy_link_events, l);
-
-			if (link) {
-				l->endpoint_link = link;
-				spa_list_append(&link->link_list, &l->link);
-			}
-
-			outport = spa_list_next(outport, link);
-			inport = spa_list_next(inport, link);
+		if (link) {
+			l->endpoint_link = link;
+			spa_list_append(&link->link_list, &l->link);
 		} else {
-			if (outport->direction != PW_DIRECTION_OUTPUT)
-				outport = spa_list_next(outport, link);
-			if (inport->direction != PW_DIRECTION_INPUT)
-				inport = spa_list_next(inport, link);
+			spa_list_append(&impl->link_list, &l->link);
 		}
 	}
 	pw_properties_free(props);
@@ -1470,39 +1627,30 @@ int sm_media_session_create_links(struct sm_media_session *sess,
 
 	/* find output node */
 	if ((str = spa_dict_lookup(dict, PW_KEY_LINK_OUTPUT_NODE)) != NULL &&
-	    (obj = find_object(impl, atoi(str))) != NULL &&
-	    strcmp(obj->type, PW_TYPE_INTERFACE_Node) == 0) {
+	    (obj = find_object(impl, atoi(str), PW_TYPE_INTERFACE_Node)) != NULL)
 		outnode = (struct sm_node*)obj;
-	}
 
 	/* find input node */
 	if ((str = spa_dict_lookup(dict, PW_KEY_LINK_INPUT_NODE)) != NULL &&
-	    (obj = find_object(impl, atoi(str))) != NULL &&
-	    strcmp(obj->type, PW_TYPE_INTERFACE_Node) == 0) {
+	    (obj = find_object(impl, atoi(str), PW_TYPE_INTERFACE_Node)) != NULL)
 		innode = (struct sm_node*)obj;
-	}
 
 	/* find endpoints and streams */
 	if ((str = spa_dict_lookup(dict, PW_KEY_ENDPOINT_LINK_OUTPUT_ENDPOINT)) != NULL &&
-	    (obj = find_object(impl, atoi(str))) != NULL &&
-	    strcmp(obj->type, PW_TYPE_INTERFACE_Endpoint) == 0) {
+	    (obj = find_object(impl, atoi(str), PW_TYPE_INTERFACE_Endpoint)) != NULL)
 		outendpoint = (struct sm_endpoint*)obj;
-	}
+
 	if ((str = spa_dict_lookup(dict, PW_KEY_ENDPOINT_LINK_OUTPUT_STREAM)) != NULL &&
-	    (obj = find_object(impl, atoi(str))) != NULL &&
-	    strcmp(obj->type, PW_TYPE_INTERFACE_EndpointStream) == 0) {
+	    (obj = find_object(impl, atoi(str), PW_TYPE_INTERFACE_EndpointStream)) != NULL)
 		outstream = (struct sm_endpoint_stream*)obj;
-	}
+
 	if ((str = spa_dict_lookup(dict, PW_KEY_ENDPOINT_LINK_INPUT_ENDPOINT)) != NULL &&
-	    (obj = find_object(impl, atoi(str))) != NULL &&
-	    strcmp(obj->type, PW_TYPE_INTERFACE_Endpoint) == 0) {
+	    (obj = find_object(impl, atoi(str), PW_TYPE_INTERFACE_Endpoint)) != NULL)
 		inendpoint = (struct sm_endpoint*)obj;
-	}
+
 	if ((str = spa_dict_lookup(dict, PW_KEY_ENDPOINT_LINK_INPUT_STREAM)) != NULL &&
-	    (obj = find_object(impl, atoi(str))) != NULL &&
-	    strcmp(obj->type, PW_TYPE_INTERFACE_EndpointStream) == 0) {
+	    (obj = find_object(impl, atoi(str), PW_TYPE_INTERFACE_EndpointStream)) != NULL)
 		instream = (struct sm_endpoint_stream*)obj;
-	}
 
 	if (outendpoint != NULL && inendpoint != NULL) {
 		link = calloc(1, sizeof(struct endpoint_link));
@@ -1545,6 +1693,36 @@ int sm_media_session_create_links(struct sm_media_session *sess,
 	return res;
 }
 
+int sm_media_session_remove_links(struct sm_media_session *sess,
+		const struct spa_dict *dict)
+{
+	struct impl *impl = SPA_CONTAINER_OF(sess, struct impl, this);
+	struct sm_object *obj;
+	struct sm_node *outnode = NULL, *innode = NULL;
+	const char *str;
+	struct link *l, *t;
+
+	/* find output node */
+	if ((str = spa_dict_lookup(dict, PW_KEY_LINK_OUTPUT_NODE)) != NULL &&
+	    (obj = find_object(impl, atoi(str), PW_TYPE_INTERFACE_Node)) != NULL)
+		outnode = (struct sm_node*)obj;
+
+	/* find input node */
+	if ((str = spa_dict_lookup(dict, PW_KEY_LINK_INPUT_NODE)) != NULL &&
+	    (obj = find_object(impl, atoi(str), PW_TYPE_INTERFACE_Node)) != NULL)
+		innode = (struct sm_node*)obj;
+
+	if (innode == NULL || outnode == NULL)
+		return -EINVAL;
+
+	spa_list_for_each_safe(l, t, &impl->link_list, link) {
+		if (l->output_node == outnode->obj.id && l->input_node == innode->obj.id) {
+			pw_proxy_destroy(l->proxy);
+		}
+	}
+	return 0;
+}
+
 static void monitor_core_done(void *data, uint32_t id, int seq)
 {
 	struct impl *impl = data;
@@ -1621,7 +1799,7 @@ static void core_error(void *data, uint32_t id, int seq, int res, const char *me
 	pw_log_error("error id:%u seq:%d res:%d (%s): %s",
 			id, seq, res, spa_strerror(res), message);
 
-	if (id == 0) {
+	if (id == PW_ID_CORE) {
 		if (res == -EPIPE)
 			pw_main_loop_quit(impl->loop);
 	}
@@ -1675,8 +1853,10 @@ static void session_shutdown(struct impl *impl)
 {
 	struct sm_object *obj;
 
-	spa_list_for_each(obj, &impl->global_list, link)
-		sm_media_session_emit_remove(impl, obj);
+	pw_log_info(NAME" %p", impl);
+
+	spa_list_consume(obj, &impl->global_list, link)
+		sm_object_destroy(obj);
 
 	sm_media_session_emit_destroy(impl);
 
@@ -1690,19 +1870,132 @@ static void session_shutdown(struct impl *impl)
 		pw_core_info_free(impl->this.info);
 }
 
+static void do_quit(void *data, int signal_number)
+{
+	struct impl *impl = data;
+	pw_main_loop_quit(impl->loop);
+}
+
+#define DEFAULT_ENABLED		"flatpak,portal,metadata,alsa-acp,alsa-seq,v4l2,bluez5,suspend-node,policy-node"
+#define DEFAULT_DISABLED	""
+
+static const struct {
+	const char *name;
+	const char *desc;
+	int (*start)(struct sm_media_session *sess);
+	const char *props;
+
+} modules[] = {
+	{ "flatpak", "manage flatpak access", sm_access_flatpak_start, NULL },
+	{ "portal", "manage portal permissions", sm_access_portal_start, NULL },
+	{ "metadata", "export metadata API", sm_metadata_start, NULL },
+	{ "alsa-seq", "alsa seq midi support", sm_alsa_midi_start, NULL },
+	{ "alsa-pcm", "alsa pcm udev detection", sm_alsa_monitor_start, NULL },
+	{ "alsa-acp", "alsa card profile udev detection", sm_alsa_monitor_start, "alsa.use-acp=true" },
+	{ "v4l2", "video for linux udev detection", sm_v4l2_monitor_start, NULL },
+	{ "libcamera", "libcamera udev detection", sm_libcamera_monitor_start, NULL },
+	{ "bluez5", "bluetooth support", sm_bluez5_monitor_start, NULL },
+	{ "suspend-node", "suspend inactive nodes", sm_suspend_node_start, NULL },
+	{ "policy-node", "configure and link nodes", sm_policy_node_start, NULL },
+};
+
+static int opt_contains(const char *opt, const char *val)
+{
+	const char *s, *state = NULL;
+	size_t len;
+	while((s = pw_split_walk(opt, ",", &len, &state)) != NULL) {
+		if (strncmp(val, s, len) == 0)
+			return 1;
+	}
+	return 0;
+}
+
+static void show_help(const char *name)
+{
+	size_t i;
+
+        fprintf(stdout, "%s [options]\n"
+             "  -h, --help                            Show this help\n"
+             "      --version                         Show version\n"
+             "  -e, --enabled                         Enabled options (default '%s')\n"
+             "  -d, --disabled                        Disabled options (default '%s')\n"
+             "  -p, --properties                      Extra properties as 'key=value { key=value }'\n",
+	     name, DEFAULT_ENABLED, DEFAULT_DISABLED);
+
+        fprintf(stdout,
+             "\noptions: (*=enabled)\n");
+	for (i = 0; i < SPA_N_ELEMENTS(modules); i++) {
+		fprintf(stdout, "\t  %c %-15.15s: %s\n",
+				opt_contains(DEFAULT_ENABLED, modules[i].name) &&
+				!opt_contains(DEFAULT_DISABLED, modules[i].name) ? '*' : ' ',
+				modules[i].name, modules[i].desc);
+	}
+}
+
 int main(int argc, char *argv[])
 {
 	struct impl impl = { 0, };
 	const struct spa_support *support;
 	uint32_t n_support;
-	int res = 0;
+	int res = 0, c;
+	const char *opt_enabled = DEFAULT_ENABLED;
+	const char *opt_disabled = DEFAULT_DISABLED;
+	const char *opt_properties = NULL;
+	static const struct option long_options[] = {
+		{ "help",	no_argument,		NULL, 'h' },
+		{ "version",	no_argument,		NULL, 'V' },
+		{ "enabled",	required_argument,	NULL, 'e' },
+		{ "disabled",	required_argument,	NULL, 'd' },
+		{ "properties",	required_argument,	NULL, 'p' },
+		{ NULL, 0, NULL, 0}
+	};
+        size_t i;
+	const struct spa_dict_item *item;
 
 	pw_init(&argc, &argv);
 
+	while ((c = getopt_long(argc, argv, "hVe:d:p:", long_options, NULL)) != -1) {
+		switch (c) {
+		case 'h':
+			show_help(argv[0]);
+			return 0;
+		case 'V':
+			fprintf(stdout, "%s\n"
+				"Compiled with libpipewire %s\n"
+				"Linked with libpipewire %s\n",
+				argv[0],
+				pw_get_headers_version(),
+				pw_get_library_version());
+			return 0;
+		case 'e':
+			opt_enabled = optarg;
+			break;
+		case 'd':
+			opt_disabled = optarg;
+			break;
+		case 'p':
+			opt_properties = optarg;
+			break;
+		default:
+			return -1;
+		}
+	}
+
+	impl.this.props = pw_properties_new_string(opt_properties ? opt_properties : "");
+	if (impl.this.props == NULL)
+		return -1;
+
+	spa_dict_for_each(item, &impl.this.props->dict)
+		pw_log_info("  '%s' = '%s'", item->key, item->value);
+
 	impl.loop = pw_main_loop_new(NULL);
 	if (impl.loop == NULL)
 		return -1;
 	impl.this.loop = pw_main_loop_get_loop(impl.loop);
+
+	pw_loop_add_signal(impl.this.loop, SIGINT, do_quit, &impl);
+	pw_loop_add_signal(impl.this.loop, SIGTERM, do_quit, &impl);
+
 	impl.this.context = pw_context_new(impl.this.loop, NULL, 0);
 	if (impl.this.context == NULL)
 		return -1;
@@ -1710,11 +2003,13 @@ int main(int argc, char *argv[])
 	pw_context_add_spa_lib(impl.this.context, "api.bluez5.*", "bluez5/libspa-bluez5");
 	pw_context_add_spa_lib(impl.this.context, "api.alsa.*", "alsa/libspa-alsa");
 	pw_context_add_spa_lib(impl.this.context, "api.v4l2.*", "v4l2/libspa-v4l2");
+	pw_context_add_spa_lib(impl.this.context, "api.libcamera.*", "libcamera/libspa-libcamera");
 
 	pw_context_set_object(impl.this.context, SM_TYPE_MEDIA_SESSION, &impl);
 
 	pw_map_init(&impl.globals, 64, 64);
 	spa_list_init(&impl.global_list);
+	spa_list_init(&impl.link_list);
 	pw_map_init(&impl.endpoint_links, 64, 64);
 	spa_list_init(&impl.endpoint_link_list);
 	spa_list_init(&impl.sync_list);
@@ -1724,7 +2019,7 @@ int main(int argc, char *argv[])
 
 	impl.dbus = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_DBus);
 	if (impl.dbus)
-		impl.this.dbus_connection = spa_dbus_get_connection(impl.dbus, DBUS_BUS_SESSION);
+		impl.this.dbus_connection = spa_dbus_get_connection(impl.dbus, SPA_DBUS_TYPE_SESSION);
 	if (impl.this.dbus_connection == NULL)
 		pw_log_warn("no dbus connection");
 	else
@@ -1735,14 +2030,22 @@ int main(int argc, char *argv[])
 	if ((res = start_policy(&impl)) < 0)
 		goto exit;
 
-	sm_metadata_start(&impl.this);
-	sm_alsa_midi_start(&impl.this);
-	sm_bluez5_monitor_start(&impl.this);
-	sm_alsa_monitor_start(&impl.this);
-	sm_v4l2_monitor_start(&impl.this);
-
-	sm_suspend_node_start(&impl.this);
-	sm_policy_node_start(&impl.this);
+	for (i = 0; i < SPA_N_ELEMENTS(modules); i++) {
+		const char *name = modules[i].name;
+		if (opt_contains(opt_enabled, name) &&
+		    !opt_contains(opt_disabled, name)) {
+			if (modules[i].props) {
+				struct pw_properties *props;
+				props = pw_properties_new_string(modules[i].props);
+				if (props) {
+					pw_properties_update(impl.this.props, &props->dict);
+					pw_properties_free(props);
+				}
+			}
+			pw_log_info("enable: %s", name);
+			modules[i].start(&impl.this);
+		}
+	}
 
 //	sm_session_manager_start(&impl.this);
 
@@ -1756,6 +2059,9 @@ exit:
 
 	pw_map_clear(&impl.endpoint_links);
 	pw_map_clear(&impl.globals);
+	pw_properties_free(impl.this.props);
+
+	pw_deinit();
 
 	return res;
 }
