@@ -37,6 +37,7 @@
 #include <spa/debug/pod.h>
 
 #include "pipewire/pipewire.h"
+#include "extensions/metadata.h"
 
 #include "media-session.h"
 
@@ -51,12 +52,18 @@ struct impl {
 	struct sm_media_session *session;
 	struct spa_hook listener;
 
+	struct spa_hook meta_listener;
+
 	struct pw_context *context;
 
 	uint32_t sample_rate;
 
 	struct spa_list node_list;
 	int seq;
+
+	uint32_t default_audio_sink;
+	uint32_t default_audio_source;
+	uint32_t default_video_source;
 };
 
 struct node {
@@ -87,19 +94,16 @@ struct node {
 	unsigned int active:1;
 	unsigned int exclusive:1;
 	unsigned int enabled:1;
+	unsigned int configured:1;
+	unsigned int dont_remix:1;
+	unsigned int monitor:1;
 };
 
-static int activate_node(struct node *node)
+static bool find_format(struct node *node)
 {
-	struct impl *impl = node->impl;
 	struct sm_param *p;
-	char buf[1024];
-	struct spa_pod_builder b = { 0, };
-	struct spa_pod *param;
 	bool have_format = false;
 
-	pw_log_debug(NAME" %p: node %p activate", impl, node);
-
 	spa_list_for_each(p, &node->obj->param_list, link) {
 		struct spa_audio_info info = { 0, };
 
@@ -125,27 +129,43 @@ static int activate_node(struct node *node)
 
 		have_format = true;
 	}
+	return have_format;
+}
 
-	if (have_format) {
-		node->format.info.raw.rate = impl->sample_rate;
+static int configure_node(struct node *node, struct spa_audio_info *info)
+{
+	struct impl *impl = node->impl;
+	char buf[1024];
+	struct spa_pod_builder b = { 0, };
+	struct spa_pod *param;
 
-		spa_pod_builder_init(&b, buf, sizeof(buf));
-		param = spa_format_audio_raw_build(&b, SPA_PARAM_Format, &node->format.info.raw);
-		param = spa_pod_builder_add_object(&b,
-			SPA_TYPE_OBJECT_ParamPortConfig, SPA_PARAM_PortConfig,
-			SPA_PARAM_PORT_CONFIG_direction, SPA_POD_Id(node->direction),
-			SPA_PARAM_PORT_CONFIG_mode,	 SPA_POD_Id(SPA_PARAM_PORT_CONFIG_MODE_dsp),
-			SPA_PARAM_PORT_CONFIG_monitor,   SPA_POD_Bool(true),
-			SPA_PARAM_PORT_CONFIG_format,    SPA_POD_Pod(param));
+	if (node->configured)
+		return 0;
 
-		if (pw_log_level_enabled(SPA_LOG_LEVEL_DEBUG))
-			spa_debug_pod(2, NULL, param);
-
-		pw_node_set_param((struct pw_node*)node->obj->obj.proxy,
-				SPA_PARAM_PortConfig, 0, param);
+	if (info != NULL) {
+		if (info->info.raw.channels < node->format.info.raw.channels || node->monitor)
+			node->format = *info;
 	}
 
-	node->active = true;
+	node->format.info.raw.rate = impl->sample_rate;
+
+	spa_pod_builder_init(&b, buf, sizeof(buf));
+	param = spa_format_audio_raw_build(&b, SPA_PARAM_Format, &node->format.info.raw);
+	param = spa_pod_builder_add_object(&b,
+		SPA_TYPE_OBJECT_ParamPortConfig, SPA_PARAM_PortConfig,
+		SPA_PARAM_PORT_CONFIG_direction, SPA_POD_Id(node->direction),
+		SPA_PARAM_PORT_CONFIG_mode,	 SPA_POD_Id(SPA_PARAM_PORT_CONFIG_MODE_dsp),
+		SPA_PARAM_PORT_CONFIG_monitor,   SPA_POD_Bool(true),
+		SPA_PARAM_PORT_CONFIG_format,    SPA_POD_Pod(param));
+
+	if (pw_log_level_enabled(SPA_LOG_LEVEL_DEBUG))
+		spa_debug_pod(2, NULL, param);
+
+	pw_node_set_param((struct pw_node*)node->obj->obj.proxy,
+			SPA_PARAM_PortConfig, 0, param);
+
+	node->configured = true;
+
 	return 0;
 }
 
@@ -157,8 +177,14 @@ static void object_update(void *data)
 	pw_log_debug(NAME" %p: node %p %08x", impl, node, node->obj->obj.changed);
 
 	if (node->obj->obj.avail & SM_NODE_CHANGE_MASK_PARAMS &&
-	    !node->active)
-		activate_node(node);
+	    !node->active) {
+		if (!find_format(node)) {
+			pw_log_debug(NAME" %p: can't find format %p", impl, node);
+			return;
+		}
+		node->active = true;
+		sm_media_session_schedule_rescan(impl->session);
+	}
 }
 
 static const struct sm_object_events object_events = {
@@ -169,7 +195,7 @@ static const struct sm_object_events object_events = {
 static int
 handle_node(struct impl *impl, struct sm_object *object)
 {
-	const char *str, *media_class;
+	const char *str, *media_class = NULL, *role;
 	enum pw_direction direction;
 	struct node *node;
 	uint32_t client_id = SPA_ID_INVALID;
@@ -177,9 +203,10 @@ handle_node(struct impl *impl, struct sm_object *object)
 	if (object->props) {
 		if ((str = pw_properties_get(object->props, PW_KEY_CLIENT_ID)) != NULL)
 			client_id = atoi(str);
-	}
 
-	media_class = object->props ? pw_properties_get(object->props, PW_KEY_MEDIA_CLASS) : NULL;
+		media_class = pw_properties_get(object->props, PW_KEY_MEDIA_CLASS);
+		role = pw_properties_get(object->props, PW_KEY_MEDIA_ROLE);
+	}
 
 	pw_log_debug(NAME" %p: node "PW_KEY_MEDIA_CLASS" %s", impl, media_class);
 
@@ -194,6 +221,9 @@ handle_node(struct impl *impl, struct sm_object *object)
 	node->type = NODE_TYPE_UNKNOWN;
 	spa_list_append(&impl->node_list, &node->link);
 
+	if (role && !strcmp(role, "DSP"))
+		node->active = node->configured = true;
+
 	if (strstr(media_class, "Stream/") == media_class) {
 		media_class += strlen("Stream/");
 
@@ -215,7 +245,10 @@ handle_node(struct impl *impl, struct sm_object *object)
 				else
 					node->plugged = SPA_TIMESPEC_TO_NSEC(&impl->now);
 			}
-			node->active = true;
+			node->active = node->configured = true;
+		}
+		else if (strstr(media_class, "Unknown") == media_class) {
+			node->active = node->configured = true;
 		}
 
 		node->direction = direction;
@@ -232,7 +265,7 @@ handle_node(struct impl *impl, struct sm_object *object)
 		else if (strstr(media_class, "Video/") == media_class) {
 			media_class += strlen("Video/");
 			media = "Video";
-			node->active = true;
+			node->active = node->configured = true;
 		}
 		else
 			return 0;
@@ -275,7 +308,7 @@ static void destroy_node(struct impl *impl, struct node *node)
 	if (node->enabled)
 		spa_hook_remove(&node->listener);
 	free(node->media);
-	if (node->peer)
+	if (node->peer && node->peer->peer == node)
 		node->peer->peer = NULL;
 	sm_object_remove_data((struct sm_object*)node->obj, SESSION_KEY);
 }
@@ -294,8 +327,7 @@ static void session_create(void *data, struct sm_object *object)
 
 	if (res < 0) {
 		pw_log_warn(NAME" %p: can't handle global %d", impl, object->id);
-	}
-	else
+	} else
 		sm_media_session_schedule_rescan(impl->session);
 }
 
@@ -314,6 +346,12 @@ static void session_remove(void *data, struct sm_object *object)
 			if (n->peer == node)
 				n->peer = NULL;
 		}
+		if (impl->default_audio_sink == object->id)
+			impl->default_audio_sink = SPA_ID_INVALID;
+		if (impl->default_audio_source == object->id)
+			impl->default_audio_source = SPA_ID_INVALID;
+		if (impl->default_video_source == object->id)
+			impl->default_video_source = SPA_ID_INVALID;
 	}
 
 	sm_media_session_schedule_rescan(impl->session);
@@ -355,10 +393,24 @@ static int find_node(void *data, struct node *node)
 		pw_log_debug(".. incompatible media %s <-> %s", node->media, find->target->media);
 		return 0;
 	}
-
 	plugged = node->plugged;
 	priority = node->priority;
 
+	if (node->media) {
+		bool is_default = false;
+		if (strcmp(node->media, "Audio") == 0) {
+			if (node->direction == PW_DIRECTION_INPUT)
+				is_default = impl->default_audio_sink == node->id;
+			else if (node->direction == PW_DIRECTION_OUTPUT)
+				is_default = impl->default_audio_source == node->id;
+		} else if (strcmp(node->media, "Video") == 0) {
+			if (node->direction == PW_DIRECTION_OUTPUT)
+				is_default = impl->default_video_source == node->id;
+		}
+		if (is_default)
+			priority += 1000;
+	}
+
 	if ((find->exclusive && node->obj->info->state == PW_NODE_STATE_RUNNING) ||
 	    (node->peer && node->peer->exclusive)) {
 		pw_log_debug(NAME " %p: node '%d' in use", impl, node->id);
@@ -386,9 +438,57 @@ static int link_nodes(struct node *node, struct node *peer)
 
 	pw_log_debug(NAME " %p: link nodes %d %d", impl, node->id, peer->id);
 
-	peer->peer = node;
+	if (node->dont_remix)
+		configure_node(node, NULL);
+	else {
+#if 0
+		bool configured = node->configured;
+		if (configured) {
+			node->configured = false;
+			pw_node_send_command((struct pw_node*)node->obj->obj.proxy,
+					&SPA_NODE_COMMAND_INIT(SPA_NODE_COMMAND_Suspend));
+		}
+#endif
+		configure_node(node, &peer->format);
+#if 0
+		if (configured) {
+			pw_node_send_command((struct pw_node*)node->obj->obj.proxy,
+					&SPA_NODE_COMMAND_INIT(SPA_NODE_COMMAND_Pause));
+		}
+#endif
+	}
+
 	node->peer = peer;
 
+	if (node->direction == PW_DIRECTION_INPUT) {
+		struct node *t = node;
+		node = peer;
+		peer = t;
+	}
+
+	props = pw_properties_new(NULL, NULL);
+	pw_properties_setf(props, PW_KEY_LINK_OUTPUT_NODE, "%d", node->id);
+	pw_properties_setf(props, PW_KEY_LINK_INPUT_NODE, "%d", peer->id);
+	pw_log_info("linking node %d to node %d", node->id, peer->id);
+
+	sm_media_session_create_links(impl->session, &props->dict);
+
+	pw_properties_free(props);
+
+	return 0;
+}
+
+static int unlink_nodes(struct node *node, struct node *peer)
+{
+	struct impl *impl = node->impl;
+	struct pw_properties *props;
+
+	pw_log_debug(NAME " %p: unlink nodes %d %d", impl, node->id, peer->id);
+
+	if (peer->peer == node)
+		peer->peer = NULL;
+	node->peer = NULL;
+
 	if (node->direction == PW_DIRECTION_INPUT) {
 		struct node *t = node;
 		node = peer;
@@ -397,10 +497,9 @@ static int link_nodes(struct node *node, struct node *peer)
 	props = pw_properties_new(NULL, NULL);
 	pw_properties_setf(props, PW_KEY_LINK_OUTPUT_NODE, "%d", node->id);
 	pw_properties_setf(props, PW_KEY_LINK_INPUT_NODE, "%d", peer->id);
-	pw_log_debug(NAME " %p: node %d -> node %d", impl,
-			node->id, peer->id);
+	pw_log_info("unlinking node %d from peer node %d", node->id, peer->id);
 
-	sm_media_session_create_links(impl->session, &props->dict);
+	sm_media_session_remove_links(impl->session, &props->dict);
 
 	pw_properties_free(props);
 
@@ -410,15 +509,17 @@ static int link_nodes(struct node *node, struct node *peer)
 static int rescan_node(struct impl *impl, struct node *n)
 {
 	struct spa_dict *props;
-        const char *str;
-        bool exclusive;
-        struct find_data find;
+	const char *str;
+	bool exclusive, reconnect;
+	struct find_data find;
 	struct pw_node_info *info;
 	struct node *peer;
 	struct sm_object *obj;
 
-	if (n->type == NODE_TYPE_DEVICE)
+	if (n->type == NODE_TYPE_DEVICE) {
+		configure_node(n, NULL);
 		return 0;
+	}
 
 	if (!n->active) {
 		pw_log_debug(NAME " %p: node %d is not active", impl, n->id);
@@ -438,10 +539,17 @@ static int rescan_node(struct impl *impl, struct node *n)
 	info = n->obj->info;
 	props = info->props;
 
+        if ((str = spa_dict_lookup(props, PW_KEY_STREAM_DONT_REMIX)) != NULL)
+		n->dont_remix = pw_properties_parse_bool(str);
+
+        if ((str = spa_dict_lookup(props, PW_KEY_STREAM_MONITOR)) != NULL)
+		n->monitor = pw_properties_parse_bool(str);
+
         str = spa_dict_lookup(props, PW_KEY_NODE_AUTOCONNECT);
         if (str == NULL || !pw_properties_parse_bool(str)) {
 		pw_log_debug(NAME" %p: node %d does not need autoconnect", impl, n->id);
-                return 0;
+		configure_node(n, NULL);
+		return 0;
 	}
 
 	if (n->media == NULL) {
@@ -460,14 +568,22 @@ static int rescan_node(struct impl *impl, struct node *n)
 	find.target = n;
 	find.exclusive = exclusive;
 
-	pw_log_info(NAME " %p: exclusive:%d", impl, exclusive);
+	pw_log_debug(NAME " %p: exclusive:%d", impl, exclusive);
+
+	str = spa_dict_lookup(props, PW_KEY_NODE_DONT_RECONNECT);
+	reconnect = str ? !pw_properties_parse_bool(str) : true;
 
 	str = spa_dict_lookup(props, PW_KEY_NODE_TARGET);
+	pw_log_info("trying to link node %d exclusive:%d reconnect:%d target:%s", n->id,
+			exclusive, reconnect, str);
+
 	if (str != NULL) {
 		uint32_t path_id = atoi(str);
-		pw_log_info(NAME " %p: target:%d", impl, path_id);
+		pw_log_debug(NAME " %p: target:%d", impl, path_id);
 
 		if ((obj = sm_media_session_find_object(impl->session, path_id)) != NULL) {
+			pw_log_debug(NAME " %p: found target:%d type:%s", impl,
+					path_id, obj->type);
 			if (strcmp(obj->type, PW_TYPE_INTERFACE_Node) == 0) {
 				peer = sm_object_get_data(obj, SESSION_KEY);
 				if (peer == NULL)
@@ -475,22 +591,27 @@ static int rescan_node(struct impl *impl, struct node *n)
 				goto do_link;
 			}
 		}
+		pw_log_warn("node %d target:%d not found, find fallback:%d", n->id,
+				path_id, reconnect);
 	}
 
-	spa_list_for_each(peer, &impl->node_list, link)
-		find_node(&find, peer);
+	if (str == NULL || reconnect) {
+		spa_list_for_each(peer, &impl->node_list, link)
+			find_node(&find, peer);
+	}
 
 	if (find.node == NULL) {
 		struct sm_object *obj;
 
-		pw_log_warn(NAME " %p: no node found for %d", impl, n->id);
+		pw_log_warn("no node found for %d", n->id);
 
-		str = spa_dict_lookup(props, PW_KEY_NODE_DONT_RECONNECT);
-		if (str != NULL && pw_properties_parse_bool(str)) {
-//			pw_registry_destroy(impl->registry, n->id);
-		}
+		if (!reconnect)
+			sm_media_session_destroy_object(impl->session, n->id);
 
 		obj = sm_media_session_find_object(impl->session, n->client_id);
+		pw_log_debug(NAME " %p: client_id:%d object:%p type:%s", impl,
+				n->client_id, obj, obj ? obj->type : "None");
+
 		if (obj && strcmp(obj->type, PW_TYPE_INTERFACE_Client) == 0) {
 			pw_client_error((struct pw_client*)obj->proxy,
 				n->id, -ENOENT, "no node available");
@@ -500,7 +621,7 @@ static int rescan_node(struct impl *impl, struct node *n)
 	peer = find.node;
 
 	if (exclusive && peer->obj->info->state == PW_NODE_STATE_RUNNING) {
-		pw_log_warn(NAME" %p: node %d busy, can't get exclusive access", impl, peer->id);
+		pw_log_warn("node %d busy, can't get exclusive access", peer->id);
 		return -EBUSY;
 	}
 	n->exclusive = exclusive;
@@ -541,6 +662,8 @@ static void session_destroy(void *data)
 {
 	struct impl *impl = data;
 	spa_hook_remove(&impl->listener);
+	if (impl->session->metadata)
+		spa_hook_remove(&impl->meta_listener);
 	free(impl);
 }
 
@@ -553,6 +676,125 @@ static const struct sm_media_session_events session_events = {
 	.destroy = session_destroy,
 };
 
+static struct node *find_node_by_id(struct impl *impl, uint32_t id)
+{
+	struct node *node;
+	spa_list_for_each(node, &impl->node_list, link) {
+		if (node->id == id)
+			return node;
+	}
+	return NULL;
+}
+
+static int move_node(struct impl *impl, uint32_t source, uint32_t target)
+{
+	struct node *n, *src_node, *dst_node;
+	const char *str;
+
+	/* find source and dest node */
+	if ((src_node = find_node_by_id(impl, source)) == NULL)
+		return -ENOENT;
+	if ((dst_node = find_node_by_id(impl, target)) == NULL)
+		return -ENOENT;
+
+	if (src_node == dst_node)
+		return 0;
+
+	pw_log_info("move %d -> %d", src_node->id, dst_node->id);
+
+	/* unlink all nodes from source and link to target */
+	spa_list_for_each(n, &impl->node_list, link) {
+		struct pw_node_info *info;
+
+		if (n->peer != src_node)
+			continue;
+
+		if ((info = n->obj->info) == NULL)
+			continue;
+
+		if ((str = spa_dict_lookup(info->props, PW_KEY_NODE_DONT_RECONNECT)) != NULL &&
+		    pw_properties_parse_bool(str))
+			continue;
+
+		unlink_nodes(n, src_node);
+		link_nodes(n, dst_node);
+	}
+	return 0;
+}
+
+static int handle_move(struct impl *impl, struct node *src_node, struct node *dst_node)
+{
+	const char *str;
+	struct pw_node_info *info;
+
+
+
+	if (src_node->peer == dst_node)
+		return 0;
+
+	if ((info = src_node->obj->info) == NULL)
+		return -EIO;
+
+	if ((str = spa_dict_lookup(info->props, PW_KEY_NODE_DONT_RECONNECT)) != NULL &&
+		    pw_properties_parse_bool(str)) {
+		pw_log_warn("can't reconnect node %d to %d", src_node->id,
+				dst_node->id);
+		return -EPERM;
+	}
+
+	pw_log_info("move node %d: from peer %d to %d", src_node->id,
+			src_node->peer ? src_node->peer->id : SPA_ID_INVALID,
+			dst_node->id);
+
+	if (src_node->peer)
+		unlink_nodes(src_node, src_node->peer);
+	link_nodes(src_node, dst_node);
+	return 0;
+}
+
+static int metadata_property(void *object, uint32_t subject,
+		const char *key, const char *type, const char *value)
+{
+	struct impl *impl = object;
+
+	if (key == NULL)
+		return 0;
+
+	if (subject == PW_ID_CORE) {
+		if (strcmp(key, "default.audio.sink") == 0) {
+			if (impl->default_audio_sink != SPA_ID_INVALID && value)
+				move_node(impl, impl->default_audio_sink, atoi(value));
+			impl->default_audio_sink = value ? (uint32_t)atoi(value) : SPA_ID_INVALID;
+		}
+		else if (strcmp(key, "default.audio.source") == 0) {
+			if (impl->default_audio_source != SPA_ID_INVALID && value)
+				move_node(impl, impl->default_audio_source, atoi(value));
+			impl->default_audio_source = value ? (uint32_t)atoi(value) : SPA_ID_INVALID;
+		} else if (strcmp(key, "default.video.source") == 0) {
+			if (impl->default_video_source != SPA_ID_INVALID && value)
+				move_node(impl, impl->default_video_source, atoi(value));
+			impl->default_video_source = value ? (uint32_t)atoi(value) : SPA_ID_INVALID;
+		}
+	} else {
+		struct node *src_node, *dst_node = NULL;
+
+		src_node = find_node_by_id(impl, subject);
+
+		if (strcmp(key, "target.node") == 0 && value != NULL) {
+			dst_node = find_node_by_id(impl, atoi(value));
+		}
+		if (src_node && dst_node)
+			handle_move(impl, src_node, dst_node);
+	}
+
+	return 0;
+}
+
+static const struct pw_metadata_events metadata_events = {
+	PW_VERSION_METADATA_EVENTS,
+	.property = metadata_property,
+};
+
 int sm_policy_node_start(struct sm_media_session *session)
 {
 	struct impl *impl;
@@ -565,10 +807,20 @@ int sm_policy_node_start(struct sm_media_session *session)
 	impl->context = session->context;
 
 	impl->sample_rate = 48000;
+	impl->default_audio_sink = SPA_ID_INVALID;
+	impl->default_audio_source = SPA_ID_INVALID;
+	impl->default_video_source = SPA_ID_INVALID;
 
 	spa_list_init(&impl->node_list);
 
-	sm_media_session_add_listener(impl->session, &impl->listener, &session_events, impl);
+	sm_media_session_add_listener(impl->session,
+			&impl->listener,
+			&session_events, impl);
 
+	if (session->metadata) {
+		pw_metadata_add_listener(session->metadata,
+				&impl->meta_listener,
+				&metadata_events, impl);
+	}
 	return 0;
 }
