@@ -34,8 +34,7 @@
 
 #include <spa/pod/iter.h>
 #include <spa/utils/result.h>
-#include <spa/debug/pod.h>
-#include <spa/debug/types.h>
+
 
 #include "config.h"
 
@@ -51,6 +50,12 @@
 #include "modules/module-protocol-native/connection.h"
 #include "modules/module-protocol-native/defs.h"
 
+#undef spa_debug
+#define spa_debug pw_log_debug
+
+#include <spa/debug/pod.h>
+#include <spa/debug/types.h>
+
 #define NAME "protocol-native"
 
 #ifndef UNIX_PATH_MAX
@@ -130,7 +135,7 @@ struct client_data {
 static void debug_msg(const char *prefix, const struct pw_protocol_native_message *msg, bool hex)
 {
 	struct spa_pod *pod;
-	fprintf(stderr, "%s: id:%d op:%d size:%d seq:%d\n", prefix,
+	pw_log_debug("%s: id:%d op:%d size:%d seq:%d", prefix,
 			msg->id, msg->opcode, msg->size, msg->seq);
 
 	if ((pod = spa_pod_from_data(msg->data, msg->size, 0, msg->size)) != NULL)
@@ -297,6 +302,7 @@ static void client_free(void *data)
 	struct client_data *this = data;
 	struct pw_impl_client *client = this->client;
 
+	pw_log_debug(NAME" %p: free", this);
 	spa_list_remove(&this->protocol_link);
 
 	if (this->source)
@@ -713,8 +719,10 @@ error:
 	pw_proxy_notify((struct pw_proxy*)this,
 			struct pw_core_events, error, 0, 0,
 			this->recv_seq, res, "connection error");
-	pw_loop_destroy_source(loop, impl->source);
-	impl->source = NULL;
+	if (impl->source) {
+		pw_loop_destroy_source(loop, impl->source);
+		impl->source = NULL;
+	}
 }
 
 static void on_need_flush(void *data)
@@ -920,6 +928,8 @@ static void destroy_server(struct pw_protocol_server *server)
 	struct server *s = SPA_CONTAINER_OF(server, struct server, this);
 	struct client_data *data, *tmp;
 
+	pw_log_debug(NAME" %p: server %p", s->this.protocol, s);
+
 	spa_list_remove(&server->link);
 	spa_hook_remove(&s->hook);
 
@@ -968,7 +978,7 @@ static const struct spa_loop_control_hooks impl_hooks = {
 };
 
 static const char *
-get_name(const struct spa_dict *props)
+get_server_name(const struct spa_dict *props)
 {
 	const char *name = NULL;
 
@@ -977,7 +987,7 @@ get_name(const struct spa_dict *props)
 	if (name == NULL)
 		name = getenv("PIPEWIRE_CORE");
 	if (name == NULL)
-		name = "pipewire-0";
+		name = PW_DEFAULT_REMOTE;
 	return name;
 }
 
@@ -1005,7 +1015,7 @@ create_server(struct pw_protocol *protocol,
 
 	pw_loop_add_hook(pw_context_get_main_loop(context), &s->hook, &impl_hooks, s);
 
-	pw_log_info(NAME" %p: created server %p", protocol, this);
+	pw_log_debug(NAME" %p: created server %p", protocol, this);
 
 	return s;
 }
@@ -1025,7 +1035,7 @@ impl_add_server(struct pw_protocol *protocol,
 
 	this = &s->this;
 
-	name = get_name(props);
+	name = get_server_name(props);
 
 	if ((res = init_socket_name(s, name)) < 0)
 		goto error;
@@ -1133,11 +1143,12 @@ static const struct pw_impl_module_events module_events = {
 
 static int need_server(struct pw_context *context, const struct spa_dict *props)
 {
-	const char *val;
+	const char *val = NULL;
 
-	val = getenv("PIPEWIRE_DAEMON");
-	if (val == NULL)
+	if (props)
 		val = spa_dict_lookup(props, PW_KEY_CORE_DAEMON);
+	if (val == NULL)
+		val = getenv("PIPEWIRE_DAEMON");
 	if (val && pw_properties_parse_bool(val))
 		return 1;
 	return 0;
