@@ -61,6 +61,7 @@
 struct impl {
 	struct pw_context this;
 	struct spa_handle *dbus_handle;
+	unsigned int recalc;
 };
 
 
@@ -357,9 +358,6 @@ void pw_context_destroy(struct pw_context *context)
 	spa_list_consume(core, &context->core_list, link)
 		pw_core_disconnect(core);
 
-	spa_list_consume(module, &context->module_list, link)
-		pw_impl_module_destroy(module);
-
 	spa_list_consume(node, &context->node_list, link)
 		pw_impl_node_destroy(node);
 
@@ -369,6 +367,9 @@ void pw_context_destroy(struct pw_context *context)
 	spa_list_consume(resource, &context->registry_resource_list, link)
 		pw_resource_destroy(resource);
 
+	spa_list_consume(module, &context->module_list, link)
+		pw_impl_module_destroy(module);
+
 	spa_list_consume(global, &context->global_list, link)
 		pw_global_destroy(global);
 
@@ -454,6 +455,14 @@ int pw_context_update_properties(struct pw_context *context, const struct spa_di
 	return changed;
 }
 
+static bool global_can_read(struct pw_context *context, struct pw_global *global)
+{
+	if (context->current_client &&
+	    !PW_PERM_IS_R(pw_global_get_permissions(global, context->current_client)))
+		return false;
+	return true;
+}
+
 SPA_EXPORT
 int pw_context_for_each_global(struct pw_context *context,
 			    int (*callback) (void *data, struct pw_global *global),
@@ -463,8 +472,7 @@ int pw_context_for_each_global(struct pw_context *context,
 	int res;
 
 	spa_list_for_each_safe(g, t, &context->global_list, link) {
-		if (context->current_client &&
-		    !PW_PERM_IS_R(pw_global_get_permissions(g, context->current_client)))
+		if (!global_can_read(context, g))
 			continue;
 		if ((res = callback(data, g)) != 0)
 			return res;
@@ -483,8 +491,7 @@ struct pw_global *pw_context_find_global(struct pw_context *context, uint32_t id
 		return NULL;
 	}
 
-	if (context->current_client &&
-	    !PW_PERM_IS_R(pw_global_get_permissions(global, context->current_client))) {
+	if (!global_can_read(context, global)) {
 		errno = EACCES;
 		return NULL;
 	}
@@ -527,8 +534,7 @@ struct pw_impl_port *pw_context_find_port(struct pw_context *context,
 		if (other_port->node == n)
 			continue;
 
-		if (context->current_client &&
-		    !PW_PERM_IS_R(pw_global_get_permissions(n->global, context->current_client)))
+		if (!global_can_read(context, n->global))
 			continue;
 
 		pw_log_debug(NAME" %p: node id:%d", context, n->global->id);
@@ -586,6 +592,45 @@ struct pw_impl_port *pw_context_find_port(struct pw_context *context,
 	return best;
 }
 
+SPA_PRINTF_FUNC(7, 8) int pw_context_debug_port_params(struct pw_context *this,
+		struct spa_node *node, enum spa_direction direction,
+		uint32_t port_id, uint32_t id, int err, const char *debug, ...)
+{
+	struct spa_pod_builder b = { 0 };
+	uint8_t buffer[4096];
+	uint32_t state;
+	struct spa_pod *param;
+	int res;
+	va_list args;
+
+	va_start(args, debug);
+	vsnprintf((char*)buffer, sizeof(buffer), debug, args);
+	va_end(args);
+
+	pw_log_error("params %s: %d:%d %s (%s)",
+			spa_debug_type_find_name(spa_type_param, id),
+			direction, port_id, spa_strerror(err), buffer);
+
+	if (err == -EBUSY)
+		return 0;
+
+        state = 0;
+        while (true) {
+                spa_pod_builder_init(&b, buffer, sizeof(buffer));
+                res = spa_node_port_enum_params_sync(node,
+                                       direction, port_id,
+                                       id, &state,
+                                       NULL, &param, &b);
+                if (res != 1) {
+			if (res < 0)
+				pw_log_error("  error: %s", spa_strerror(res));
+                        break;
+		}
+                pw_log_pod(SPA_LOG_LEVEL_ERROR, param);
+        }
+        return 0;
+}
+
 /** Find a common format between two ports
  *
  * \param context a context object
@@ -646,18 +691,20 @@ int pw_context_find_format(struct pw_context *context,
 			goto error;
 		}
 		pw_log_debug(NAME" %p: Got output format:", context);
-		if (pw_log_level_enabled(SPA_LOG_LEVEL_DEBUG))
-			spa_debug_format(2, NULL, filter);
+		pw_log_format(SPA_LOG_LEVEL_DEBUG, filter);
 
 		if ((res = spa_node_port_enum_params_sync(input->node->node,
 						     input->direction, input->port_id,
 						     SPA_PARAM_EnumFormat, &iidx,
 						     filter, format, builder)) <= 0) {
-			if (res < 0)
+			if (res == -ENOENT || res == 0) {
+				pw_log_debug(NAME" %p: no input format filter, using output format: %s",
+						context, spa_strerror(res));
+				*format = filter;
+			} else {
 				*error = spa_aprintf("error input enum formats: %s", spa_strerror(res));
-			else
-				*error = spa_aprintf("no input formats");
-			goto error;
+				goto error;
+			}
 		}
 	} else if (out_state >= PW_IMPL_PORT_STATE_CONFIGURE && in_state > PW_IMPL_PORT_STATE_CONFIGURE) {
 		/* only output needs format */
@@ -673,18 +720,20 @@ int pw_context_find_format(struct pw_context *context,
 			goto error;
 		}
 		pw_log_debug(NAME" %p: Got input format:", context);
-		if (pw_log_level_enabled(SPA_LOG_LEVEL_DEBUG))
-			spa_debug_format(2, NULL, filter);
+		pw_log_format(SPA_LOG_LEVEL_DEBUG, filter);
 
 		if ((res = spa_node_port_enum_params_sync(output->node->node,
 						     output->direction, output->port_id,
 						     SPA_PARAM_EnumFormat, &oidx,
 						     filter, format, builder)) <= 0) {
-			if (res < 0)
+			if (res == -ENOENT || res == 0) {
+				pw_log_debug(NAME" %p: no output format filter, using input format: %s",
+						context, spa_strerror(res));
+				*format = filter;
+			} else {
 				*error = spa_aprintf("error output enum formats: %s", spa_strerror(res));
-			else
-				*error = spa_aprintf("no output format");
-			goto error;
+				goto error;
+			}
 		}
 	} else if (in_state == PW_IMPL_PORT_STATE_CONFIGURE && out_state == PW_IMPL_PORT_STATE_CONFIGURE) {
 	      again:
@@ -695,25 +744,25 @@ int pw_context_find_format(struct pw_context *context,
 						     input->direction, input->port_id,
 						     SPA_PARAM_EnumFormat, &iidx,
 						     NULL, &filter, &fb)) != 1) {
-			if (res == 0 && iidx == 0) {
-				*error = spa_aprintf("no compatible formats");
+			if (res == -ENOENT) {
+				pw_log_debug(NAME" %p: no input filter", context);
+				filter = NULL;
+			} else {
+				if (res < 0)
+					*error = spa_aprintf("error input enum formats: %s", spa_strerror(res));
+				else
+					*error = spa_aprintf("no more input formats");
 				goto error;
 			}
-			if (res < 0)
-				*error = spa_aprintf("error input enum formats: %s", spa_strerror(res));
-			else
-				*error = spa_aprintf("no more input formats");
-			goto error;
 		}
 		pw_log_debug(NAME" %p: enum output %d with filter: %p", context, oidx, filter);
-		if (pw_log_level_enabled(SPA_LOG_LEVEL_DEBUG))
-			spa_debug_format(2, NULL, filter);
+		pw_log_format(SPA_LOG_LEVEL_DEBUG, filter);
 
 		if ((res = spa_node_port_enum_params_sync(output->node->node,
 						     output->direction, output->port_id,
 						     SPA_PARAM_EnumFormat, &oidx,
 						     filter, format, builder)) != 1) {
-			if (res == 0) {
+			if (res == 0 && filter != NULL) {
 				oidx = 0;
 				goto again;
 			}
@@ -722,30 +771,35 @@ int pw_context_find_format(struct pw_context *context,
 		}
 
 		pw_log_debug(NAME" %p: Got filtered:", context);
-		if (pw_log_level_enabled(SPA_LOG_LEVEL_DEBUG))
-			spa_debug_format(2, NULL, *format);
+		pw_log_format(SPA_LOG_LEVEL_DEBUG, *format);
 	} else {
 		res = -EBADF;
 		*error = spa_aprintf("error bad node state");
 		goto error;
 	}
 	return res;
-
 error:
 	if (res == 0)
-		res = -ENOENT;
+		res = -EINVAL;
 	return res;
 }
 
+static int ensure_state(struct pw_impl_node *node, bool running)
+{
+	enum pw_node_state state = node->info.state;
+	if (node->active && running)
+		state = PW_NODE_STATE_RUNNING;
+	else if (state > PW_NODE_STATE_IDLE)
+		state = PW_NODE_STATE_IDLE;
+	return pw_impl_node_set_state(node, state);
+}
+
 static int collect_nodes(struct pw_impl_node *driver)
 {
 	struct spa_list queue;
 	struct pw_impl_node *n, *t;
 	struct pw_impl_port *p;
 	struct pw_impl_link *l;
-	uint32_t max_quantum = 0;
-	uint32_t min_quantum = 0;
-	uint32_t quantum;
 
 	spa_list_consume(t, &driver->follower_list, follower_link) {
 		spa_list_remove(&t->follower_link);
@@ -761,18 +815,18 @@ static int collect_nodes(struct pw_impl_node *driver)
 	spa_list_consume(n, &queue, sort_link) {
 		spa_list_remove(&n->sort_link);
 		pw_impl_node_set_driver(n, driver);
-
-		if (n->quantum_size > 0) {
-			if (min_quantum == 0 || n->quantum_size < min_quantum)
-				min_quantum = n->quantum_size;
-			if (n->quantum_size > max_quantum)
-				max_quantum = n->quantum_size;
-		}
+		n->passive = true;
 
 		spa_list_for_each(p, &n->input_ports, link) {
 			spa_list_for_each(l, &p->links, input_link) {
 				t = l->output->node;
-				if (!t->visited && t->active) {
+				if (!l->passive)
+					driver->passive = n->passive = false;
+				else
+					pw_impl_link_prepare(l);
+				if (t->visited || !t->active)
+					continue;
+				if (l->prepared) {
 					t->visited = true;
 					spa_list_append(&queue, &t->sort_link);
 				}
@@ -781,106 +835,145 @@ static int collect_nodes(struct pw_impl_node *driver)
 		spa_list_for_each(p, &n->output_ports, link) {
 			spa_list_for_each(l, &p->links, output_link) {
 				t = l->input->node;
-				if (!t->visited && t->active) {
+				if (!l->passive)
+					driver->passive = n->passive = false;
+				else
+					pw_impl_link_prepare(l);
+				if (t->visited || !t->active)
+					continue;
+				if (l->prepared) {
 					t->visited = true;
 					spa_list_append(&queue, &t->sort_link);
 				}
 			}
 		}
 	}
-
-	quantum = min_quantum;
-	if (quantum == 0)
-		quantum = driver->context->defaults.clock_quantum;
-
-	/* for now, we try to limit the latency between min and default, We can
-	 * go to max but we should really only do this when in power save mode */
-	driver->quantum_current = SPA_CLAMP(quantum,
-			driver->context->defaults.clock_min_quantum,
-			driver->context->defaults.clock_max_quantum);
-
 	return 0;
 }
 
-int pw_context_recalc_graph(struct pw_context *context)
+int pw_context_recalc_graph(struct pw_context *context, const char *reason)
 {
-	struct pw_impl_node *n, *s, *target;
+	struct impl *impl = SPA_CONTAINER_OF(context, struct impl, this);
+	struct pw_impl_node *n, *s, *target, *fallback;
+
+	pw_log_info(NAME" %p: busy:%d reason:%s", context, impl->recalc, reason);
+
+	if (impl->recalc)
+		return -EBUSY;
+
+	impl->recalc = true;
 
 	/* start from all drivers and group all nodes that are linked
 	 * to it. Some nodes are not (yet) linked to anything and they
 	 * will end up 'unassigned' to a master. Other nodes are master
 	 * and if they have active followers, we can use them to schedule
 	 * the unassigned nodes. */
-	target = NULL;
+	target = fallback = NULL;
 	spa_list_for_each(n, &context->driver_list, driver_link) {
-		uint32_t active_followers;
-
-		if (n->active && !n->visited)
-			collect_nodes(n);
-
-		/* from now on we are only interested in nodes that are
-		 * a master. We're going to count the number of followers it
-		 * has. */
-		if (!n->master)
+		if (n->exported)
 			continue;
 
-		active_followers = 0;
+		if (!n->visited)
+			collect_nodes(n);
+
+		/* from now on we are only interested in active master nodes.
+		 * We're going to see if there are active followers. */
+		if (!n->master || !n->active || n->passive)
+			continue;
+
+		/* first active master node is fallback */
+		if (fallback == NULL)
+			fallback = n;
+
 		spa_list_for_each(s, &n->follower_list, follower_link) {
 			pw_log_debug(NAME" %p: driver %p: follower %p %s: %d",
 					context, n, s, s->name, s->active);
-			if (s != n && s->active)
-				active_followers++;
-		}
-		pw_log_debug(NAME" %p: driver %p active followers %d",
-				context, n, active_followers);
-
-		/* if the master has active followers, it is a target for our
-		 * unassigned nodes */
-		if (active_followers > 0) {
-			if (target == NULL)
-				target = n;
+			if (s != n && s->active) {
+				/* if the master has active followers, it is a target for our
+				 * unassigned nodes */
+				if (target == NULL)
+					target = n;
+				break;
+			}
 		}
 	}
+	/* no active node, use fallback master */
+	if (target == NULL)
+		target = fallback;
 
 	/* now go through all available nodes. The ones we didn't visit
 	 * in collect_nodes() are not linked to any master. We assign them
-	 * to an active master */
+	 * to either an active master of the first master */
 	spa_list_for_each(n, &context->node_list, link) {
+		if (n->exported)
+			continue;
+
 		if (!n->visited) {
 			struct pw_impl_node *t;
 
 			pw_log_debug(NAME" %p: unassigned node %p: '%s' %d %d", context,
 					n, n->name, n->active, n->want_driver);
 
-			t = n->want_driver ? target : NULL;
+			t = n->active && n->want_driver ? target : NULL;
 
-			if (t != NULL) {
-				if (n->quantum_size > 0 && n->quantum_size < t->quantum_current)
-					t->quantum_current =
-						SPA_MAX(context->defaults.clock_min_quantum, n->quantum_size);
-			}
 			pw_impl_node_set_driver(n, t);
-			pw_impl_node_set_state(n, t && n->active ?
-					PW_NODE_STATE_RUNNING : PW_NODE_STATE_IDLE);
+			if (t == NULL)
+				ensure_state(n, false);
 		}
 		n->visited = false;
 	}
 
-	/* assign final quantum and debug masters and followers */
+	/* assign final quantum and set state for followers and master */
 	spa_list_for_each(n, &context->driver_list, driver_link) {
-		if (!n->master)
+		bool running = false;
+		uint32_t max_quantum = 0;
+		uint32_t min_quantum = 0;
+		uint32_t quantum;
+
+		if (!n->master || n->exported)
 			continue;
 
-		if (n->rt.position && n->quantum_current != n->rt.position->clock.duration)
-			n->rt.position->clock.duration = n->quantum_current;
+		/* collect quantum and count active nodes */
+		spa_list_for_each(s, &n->follower_list, follower_link) {
+			if (s == n)
+				continue;
+			if (s->active)
+				running = !n->passive;
+			if (s->quantum_size > 0) {
+				if (min_quantum == 0 || s->quantum_size < min_quantum)
+					min_quantum = s->quantum_size;
+				if (s->quantum_size > max_quantum)
+					max_quantum = s->quantum_size;
+			}
+		}
+		quantum = min_quantum;
+		if (quantum == 0)
+			quantum = context->defaults.clock_quantum;
+		quantum = SPA_CLAMP(quantum,
+				context->defaults.clock_min_quantum,
+				context->defaults.clock_max_quantum);
 
-		pw_log_debug(NAME" %p: master %p quantum:%u '%s'", context, n,
-				n->quantum_current, n->name);
+		if (n->rt.position && quantum != n->rt.position->clock.duration) {
+			pw_log_info("(%s-%u) new quantum:%"PRIu64"->%u",
+					n->name, n->info.id,
+					n->rt.position->clock.duration,
+					quantum);
+			n->rt.position->clock.duration = quantum;
+		}
 
-		spa_list_for_each(s, &n->follower_list, follower_link)
+		pw_log_debug(NAME" %p: master %p running:%d passive:%d quantum:%u '%s'", context, n,
+				running, n->passive, quantum, n->name);
+
+		spa_list_for_each(s, &n->follower_list, follower_link) {
+			if (s == n)
+				continue;
 			pw_log_debug(NAME" %p: follower %p: active:%d '%s'",
 					context, s, s->active, s->name);
+			ensure_state(s, running);
+		}
+		ensure_state(n, running);
 	}
+	impl->recalc = false;
 	return 0;
 }
 
@@ -954,6 +1047,10 @@ struct spa_handle *pw_context_load_spa_handle(struct pw_context *context,
 SPA_EXPORT
 int pw_context_register_export_type(struct pw_context *context, struct pw_export_type *type)
 {
+	if (pw_context_find_export_type(context, type->type)) {
+		pw_log_warn("context %p: duplicate export type %s", context, type->type);
+		return -EEXIST;
+	}
 	pw_log_debug("context %p: Add export type %s to context", context, type->type);
 	spa_list_append(&context->export_list, &type->link);
 	return 0;
