@@ -162,6 +162,8 @@ static int core_hello(void *object, uint32_t version)
 	pw_log_debug(NAME" %p: hello %d from resource %p", context, version, resource);
 	pw_map_for_each(&client->objects, destroy_resource, client);
 
+	pw_mempool_clear(client->pool);
+
 	this->info.change_mask = PW_CORE_CHANGE_MASK_ALL;
 	pw_core_resource_info(resource, &this->info);
 
@@ -202,7 +204,7 @@ static int core_error(void *object, uint32_t id, int seq, int res, const char *m
 	struct pw_impl_client *client = resource->client;
 	struct pw_resource *r;
 
-	pw_log_debug(NAME" %p: error %d for resource %d: %s", resource->context, res, id, message);
+	pw_log_error(NAME" %p: error %d for resource %d: %s", resource->context, res, id, message);
 
 	if ((r = pw_impl_client_find_resource(client, id)) == NULL)
 		return -EINVAL;
