@@ -116,10 +116,11 @@ static void check_properties(struct pw_impl_device *device)
 {
 	const char *str;
 
-	if ((str = pw_properties_get(device->properties, PW_KEY_DEVICE_NAME))) {
+	if ((str = pw_properties_get(device->properties, PW_KEY_DEVICE_NAME)) &&
+	    (device->name == NULL || strcmp(str, device->name) != 0)) {
 		free(device->name);
 		device->name = strdup(str);
-		pw_log_info(NAME" %p: name '%s'", device, device->name);
+		pw_log_debug(NAME" %p: name '%s'", device, device->name);
 	}
 }
 
@@ -139,6 +140,7 @@ struct pw_impl_device *pw_context_create_device(struct pw_context *context,
 	}
 
 	this = &impl->this;
+	this->name = strdup("device");
 	pw_log_debug(NAME" %p: new", this);
 
 	if (properties == NULL)
@@ -200,7 +202,22 @@ void pw_impl_device_destroy(struct pw_impl_device *device)
 	free(device);
 }
 
-static void device_pong(void *data, int seq)
+static void remove_busy_resource(struct resource_data *d)
+{
+	if (d->end != -1) {
+		spa_hook_remove(&d->listener);
+		d->end = -1;
+		pw_impl_client_set_busy(d->resource->client, false);
+	}
+}
+
+static void resource_destroy(void *data)
+{
+	struct resource_data *d = data;
+	remove_busy_resource(d);
+}
+
+static void resource_pong(void *data, int seq)
 {
 	struct resource_data *d = data;
 	struct pw_resource *resource = d->resource;
@@ -210,7 +227,8 @@ static void device_pong(void *data, int seq)
 
 static const struct pw_resource_events resource_events = {
 	PW_VERSION_RESOURCE_EVENTS,
-	.pong = device_pong,
+	.destroy = resource_destroy,
+	.pong = resource_pong,
 };
 
 static void result_device_params(void *data, int seq, int res, uint32_t type, const void *result)
@@ -278,11 +296,8 @@ static void result_device_params_async(void *data, int seq, int res, uint32_t ty
 	pw_log_debug(NAME" %p: async result %d %d (%d/%d)", d->device,
 			res, seq, d->seq, d->end);
 
-	if (seq == d->end) {
-		spa_hook_remove(&d->listener);
-		d->end = -1;
-		pw_impl_client_set_busy(d->resource->client, false);
-	}
+	if (seq == d->end)
+		remove_busy_resource(d);
 	if (seq == d->seq)
 		result_device_params(&d->data, seq, res, type, result);
 
@@ -348,11 +363,8 @@ static void result_device_done(void *data, int seq, int res, uint32_t type, cons
 	pw_log_debug(NAME" %p: async result %d %d (%d/%d)", d->device,
 			res, seq, d->seq, d->end);
 
-	if (seq == d->end) {
-		spa_hook_remove(&d->listener);
-		d->end = -1;
-		pw_impl_client_set_busy(d->resource->client, false);
-	}
+	if (seq == d->end)
+		remove_busy_resource(d);
 }
 
 static int device_set_param(void *object, uint32_t id, uint32_t flags,
@@ -732,13 +744,9 @@ static void device_object_info(void *data, uint32_t id,
 	od = find_object(device, id);
 
 	if (info == NULL) {
-		if (od) {
-			pw_log_debug(NAME" %p: remove node %d", device, id);
+		pw_log_debug(NAME" %p: remove node %d (%p)", device, id, od);
+		if (od)
 			object_destroy(od);
-		}
-		else {
-			pw_log_warn(NAME" %p: unknown node %d", device, id);
-		}
 	}
 	else if (od != NULL) {
 		if (info->change_mask & SPA_DEVICE_OBJECT_CHANGE_MASK_PROPS)
