@@ -45,6 +45,8 @@
 struct impl {
 	struct pw_impl_node this;
 
+	enum pw_node_state pending;
+
 	struct pw_work_queue *work;
 
 	int last_error;
@@ -58,8 +60,9 @@ struct impl {
 
 struct resource_data {
 	struct pw_impl_node *node;
-	struct pw_resource *resource;
 
+	struct pw_resource *resource;
+	struct spa_hook resource_listener;
 	struct spa_hook object_listener;
 
 	uint32_t subscribe_ids[MAX_PARAMS];
@@ -91,29 +94,35 @@ static void node_deactivate(struct pw_impl_node *this)
 
 static void add_node(struct pw_impl_node *this, struct pw_impl_node *driver)
 {
-	uint32_t rdriver, rnode;
+	struct pw_node_activation_state *dstate, *nstate;
 
 	if (this->exported)
 		return;
 
 	pw_log_trace(NAME" %p: add to driver %p %p %p", this, driver,
 			driver->rt.activation, this->rt.activation);
+
 	/* signal the driver */
 	this->rt.driver_target.activation = driver->rt.activation;
 	this->rt.driver_target.node = driver;
 	this->rt.driver_target.data = driver;
 	spa_list_append(&this->rt.target_list, &this->rt.driver_target.link);
-	rdriver = ++this->rt.driver_target.activation->state[0].required;
+
+	dstate = &this->rt.driver_target.activation->state[0];
+	dstate->required++;
 
 	spa_list_append(&driver->rt.target_list, &this->rt.target.link);
-	rnode = ++this->rt.activation->state[0].required;
+	nstate = &this->rt.activation->state[0];
+	nstate->required++;
 
-	pw_log_trace(NAME" %p: required driver:%d node:%d", this, rdriver, rnode);
+	pw_log_trace(NAME" %p: driver state:%p pending:%d/%d, node state:%p pending:%d/%d",
+			this, dstate, dstate->pending, dstate->required,
+			nstate, nstate->pending, nstate->required);
 }
 
 static void remove_node(struct pw_impl_node *this)
 {
-	uint32_t rdriver, rnode;
+	struct pw_node_activation_state *dstate, *nstate;
 
 	if (this->exported)
 		return;
@@ -123,12 +132,16 @@ static void remove_node(struct pw_impl_node *this)
 			this->rt.driver_target.activation, this->rt.activation);
 
 	spa_list_remove(&this->rt.driver_target.link);
-	rdriver = --this->rt.driver_target.activation->state[0].required;
+	dstate = &this->rt.driver_target.activation->state[0];
+	dstate->required--;
 
 	spa_list_remove(&this->rt.target.link);
-	rnode = --this->rt.activation->state[0].required;
+	nstate = &this->rt.activation->state[0];
+	nstate->required--;
 
-	pw_log_trace(NAME" %p: required driver:%d node:%d", this, rdriver, rnode);
+	pw_log_trace(NAME" %p: driver state:%p pending:%d/%d, node state:%p pending:%d/%d",
+			this, dstate, dstate->pending, dstate->required,
+			nstate, nstate->pending, nstate->required);
 }
 
 static int
@@ -148,10 +161,10 @@ static int pause_node(struct pw_impl_node *this)
 	struct impl *impl = SPA_CONTAINER_OF(this, struct impl, this);
 	int res = 0;
 
-	pw_log_debug(NAME" %p: pause node state:%s", this,
-			pw_node_state_as_string(this->info.state));
+	pw_log_debug(NAME" %p: pause node state:%s pause-on-idle:%d", this,
+			pw_node_state_as_string(this->info.state), impl->pause_on_idle);
 
-	if (this->info.state <= PW_NODE_STATE_IDLE && impl->pause_on_idle)
+	if (impl->pending <= PW_NODE_STATE_IDLE)
 		return 0;
 
 	node_deactivate(this);
@@ -166,41 +179,22 @@ static int pause_node(struct pw_impl_node *this)
 	return res;
 }
 
-static int
-do_node_add(struct spa_loop *loop,
-	    bool async, uint32_t seq, const void *data, size_t size, void *user_data)
-{
-	struct pw_impl_node *this = user_data;
-	struct pw_impl_node *driver = this->driver_node;
-
-	if (this->source.loop == NULL) {
-		spa_loop_add_source(loop, &this->source);
-		add_node(this, driver);
-	}
-	return 0;
-}
-
 static int start_node(struct pw_impl_node *this)
 {
+	struct impl *impl = SPA_CONTAINER_OF(this, struct impl, this);
 	int res = 0;
 
-	if (this->info.state >= PW_NODE_STATE_RUNNING)
+	if (impl->pending >= PW_NODE_STATE_RUNNING)
 		return 0;
 
-	pw_log_debug(NAME" %p: start node %d %d %d %d", this, this->n_ready_output_links,
-			this->n_used_output_links, this->n_ready_input_links,
-			this->n_used_input_links);
-
-	if (this->n_ready_output_links != this->n_used_output_links ||
-	    this->n_ready_input_links != this->n_used_input_links)
-		return 0;
+	pw_log_debug(NAME" %p: start node", this);
 
 	res = spa_node_send_command(this->node,
 				    &SPA_NODE_COMMAND_INIT(SPA_NODE_COMMAND_Start));
 
 	if (res < 0)
-		pw_log_error(NAME" %p: start node error %d: %s",
-				this, res, spa_strerror(res));
+		pw_log_error("(%s-%u) start node error %d: %s", this->name, this->info.id,
+				res, spa_strerror(res));
 
 	return res;
 }
@@ -278,25 +272,43 @@ static void emit_params(struct pw_impl_node *node, uint32_t *changed_ids, uint32
 	}
 }
 
+static int
+do_node_add(struct spa_loop *loop,
+	    bool async, uint32_t seq, const void *data, size_t size, void *user_data)
+{
+	struct pw_impl_node *this = user_data;
+	struct pw_impl_node *driver = this->driver_node;
+
+	if (this->source.loop == NULL) {
+		spa_loop_add_source(loop, &this->source);
+		add_node(this, driver);
+	}
+	return 0;
+}
+
 static void node_update_state(struct pw_impl_node *node, enum pw_node_state state, char *error)
 {
-	enum pw_node_state old;
-
-	old = node->info.state;
-	if (old == state)
-		return;
-
-	if (state == PW_NODE_STATE_ERROR) {
-		pw_log_error(NAME" %p: update state from %s -> error (%s)", node,
-		     pw_node_state_as_string(old), error);
-	} else {
-		pw_log_debug(NAME" %p: update state from %s -> %s", node,
-		     pw_node_state_as_string(old), pw_node_state_as_string(state));
-	}
+	struct impl *impl = SPA_CONTAINER_OF(node, struct impl, this);
+	enum pw_node_state old = node->info.state;
 
 	free((char*)node->info.error);
 	node->info.error = error;
 	node->info.state = state;
+	impl->pending = state;
+
+	if (old == state)
+		return;
+
+	pw_log_debug(NAME" %p: (%s) %s -> %s (%s)", node, node->name,
+		     pw_node_state_as_string(old), pw_node_state_as_string(state), error);
+
+	if (state == PW_NODE_STATE_ERROR) {
+		pw_log_error("(%s-%u) %s -> error (%s)", node->name, node->info.id,
+		     pw_node_state_as_string(old), error);
+	} else {
+		pw_log_info("(%s-%u) %s -> %s", node->name, node->info.id,
+		     pw_node_state_as_string(old), pw_node_state_as_string(state));
+	}
 
 	switch (state) {
 	case PW_NODE_STATE_RUNNING:
@@ -320,6 +332,9 @@ static int suspend_node(struct pw_impl_node *this)
 	pw_log_debug(NAME" %p: suspend node state:%s", this,
 			pw_node_state_as_string(this->info.state));
 
+	if (this->info.state <= PW_NODE_STATE_SUSPENDED)
+		return 0;
+
 	pause_node(this);
 
 	spa_list_for_each(p, &this->input_ports, link) {
@@ -340,6 +355,9 @@ static int suspend_node(struct pw_impl_node *this)
 
 	res = spa_node_send_command(this->node,
 				    &SPA_NODE_COMMAND_INIT(SPA_NODE_COMMAND_Suspend));
+	if (res == -ENOTSUP)
+		res = spa_node_send_command(this->node,
+				    &SPA_NODE_COMMAND_INIT(SPA_NODE_COMMAND_Pause));
 	if (res < 0 && res != -EIO)
 		pw_log_warn(NAME" %p: suspend node error %s", this, spa_strerror(res));
 
@@ -404,16 +422,22 @@ static int node_subscribe_params(void *object, uint32_t *ids, uint32_t n_ids)
 	return 0;
 }
 
+static void remove_busy_resource(struct resource_data *d)
+{
+	if (d->end != -1) {
+		spa_hook_remove(&d->listener);
+		d->end = -1;
+		pw_impl_client_set_busy(d->resource->client, false);
+	}
+}
+
 static void result_node_sync(void *data, int seq, int res, uint32_t type, const void *result)
 {
 	struct resource_data *d = data;
 
 	pw_log_debug(NAME" %p: sync result %d %d (%d/%d)", d->node, res, seq, d->seq, d->end);
-	if (seq == d->end) {
-		spa_hook_remove(&d->listener);
-		d->end = -1;
-		pw_impl_client_set_busy(d->resource->client, false);
-	}
+	if (seq == d->end)
+		remove_busy_resource(d);
 }
 
 static int node_set_param(void *object, uint32_t id, uint32_t flags,
@@ -474,6 +498,26 @@ static const struct pw_node_methods node_methods = {
 	.send_command = node_send_command
 };
 
+static void resource_destroy(void *data)
+{
+	struct resource_data *d = data;
+	remove_busy_resource(d);
+}
+
+static void resource_pong(void *data, int seq)
+{
+	struct resource_data *d = data;
+	struct pw_resource *resource = d->resource;
+	pw_log_debug(NAME" %p: resource %p: got pong %d", d->node,
+			resource, seq);
+}
+
+static const struct pw_resource_events resource_events = {
+	PW_VERSION_RESOURCE_EVENTS,
+	.destroy = resource_destroy,
+	.pong = resource_pong,
+};
+
 static int
 global_bind(void *_data, struct pw_impl_client *client, uint32_t permissions,
 	    uint32_t version, uint32_t id)
@@ -492,6 +536,9 @@ global_bind(void *_data, struct pw_impl_client *client, uint32_t permissions,
 	data->resource = resource;
 	data->end = -1;
 
+	pw_resource_add_listener(resource,
+			&data->resource_listener,
+			&resource_events, data);
 	pw_resource_add_object_listener(resource,
 			&data->object_listener,
 			&node_methods, data);
@@ -534,6 +581,28 @@ static inline void insert_driver(struct pw_context *context, struct pw_impl_node
 	spa_list_append(&n->driver_link, &node->driver_link);
 }
 
+static void update_io(struct pw_impl_node *node)
+{
+	pw_log_debug(NAME" %p: id:%d", node, node->info.id);
+
+	if (spa_node_set_io(node->node,
+			    SPA_IO_Position,
+			    &node->rt.activation->position,
+			    sizeof(struct spa_io_position)) >= 0) {
+		pw_log_debug(NAME" %p: set position %p", node, &node->rt.activation->position);
+		node->rt.position = &node->rt.activation->position;
+	} else if (node->driver) {
+		pw_log_warn(NAME" %p: can't set position on driver", node);
+	}
+	if (spa_node_set_io(node->node,
+			    SPA_IO_Clock,
+			    &node->rt.activation->position.clock,
+			    sizeof(struct spa_io_clock)) >= 0) {
+		pw_log_debug(NAME" %p: set clock %p", node, &node->rt.activation->position.clock);
+		node->rt.clock = &node->rt.activation->position.clock;
+	}
+}
+
 SPA_EXPORT
 int pw_impl_node_register(struct pw_impl_node *this,
 		     struct pw_properties *properties)
@@ -548,6 +617,7 @@ int pw_impl_node_register(struct pw_impl_node *this,
 		PW_KEY_DEVICE_ID,
 		PW_KEY_PRIORITY_SESSION,
 		PW_KEY_PRIORITY_MASTER,
+		PW_KEY_APP_NAME,
 		PW_KEY_NODE_DESCRIPTION,
 		PW_KEY_NODE_NAME,
 		PW_KEY_NODE_NICK,
@@ -596,12 +666,16 @@ int pw_impl_node_register(struct pw_impl_node *this,
 	pw_global_add_listener(this->global, &this->global_listener, &global_events, this);
 	pw_global_register(this->global);
 
+	if (this->node)
+		update_io(this);
+
 	spa_list_for_each(port, &this->input_ports, link)
 		pw_impl_port_register(port, NULL);
 	spa_list_for_each(port, &this->output_ports, link)
 		pw_impl_port_register(port, NULL);
 
-	pw_context_recalc_graph(context);
+	if (this->active)
+		pw_context_recalc_graph(context, "register active node");
 
 	return 0;
 
@@ -663,7 +737,11 @@ int pw_impl_node_set_driver(struct pw_impl_node *node, struct pw_impl_node *driv
 	remove_segment_master(old, node->info.id);
 
 	node->master = node->driver && driver == node;
-	pw_log_info(NAME" %p: driver %p (%s) master:%u", node, driver, driver->name, node->master);
+	pw_log_debug(NAME" %p: driver %p master:%u", node,
+		driver, node->master);
+	pw_log_info("(%s-%u) -> change driver (%s-%d -> %s-%d)",
+			node->name, node->info.id,
+			old->name, old->info.id, driver->name, driver->info.id);
 
 	node->driver_node = driver;
 
@@ -698,19 +776,20 @@ static uint32_t flp2(uint32_t x)
 static void check_properties(struct pw_impl_node *node)
 {
 	struct impl *impl = SPA_CONTAINER_OF(node, struct impl, this);
+	struct pw_context *context = node->context;
 	const char *str;
 	bool driver, do_recalc = false;
 
 	if ((str = pw_properties_get(node->properties, PW_KEY_PRIORITY_MASTER))) {
 		node->priority_master = pw_properties_parse_int(str);
-		pw_log_info(NAME" %p: priority master %d", node, node->priority_master);
+		pw_log_debug(NAME" %p: priority master %d", node, node->priority_master);
 	}
 
 	if ((str = pw_properties_get(node->properties, PW_KEY_NODE_NAME)) &&
-	    (node->name == NULL || strcmp(str, node->name) != 0)) {
+	    (node->name == NULL || strcmp(node->name, str) != 0)) {
 		free(node->name);
 		node->name = strdup(str);
-		pw_log_info(NAME" %p: name '%s'", node, node->name);
+		pw_log_debug(NAME" %p: name '%s'", node, node->name);
 	}
 
 	if ((str = pw_properties_get(node->properties, PW_KEY_NODE_PAUSE_ON_IDLE)))
@@ -729,11 +808,11 @@ static void check_properties(struct pw_impl_node *node)
 		node->want_driver = false;
 
 	if (node->driver != driver) {
-		pw_log_info(NAME" %p: driver %d -> %d", node, node->driver, driver);
+		pw_log_debug(NAME" %p: driver %d -> %d", node, node->driver, driver);
 		node->driver = driver;
 		if (node->registered) {
 			if (driver)
-				insert_driver(node->context, node);
+				insert_driver(context, node);
 			else
 				spa_list_remove(&node->driver_link);
 		}
@@ -744,11 +823,13 @@ static void check_properties(struct pw_impl_node *node)
                 if (sscanf(str, "%u/%u", &num, &denom) == 2 && denom != 0) {
 			uint32_t quantum_size;
 
-			quantum_size = flp2((num * node->context->defaults.clock_rate / denom));
+			quantum_size = flp2((num * context->defaults.clock_rate / denom));
 
 			if (quantum_size != node->quantum_size) {
-				pw_log_info(NAME" %p: latency '%s' quantum %u/%u",
-						node, str, quantum_size, node->context->defaults.clock_rate);
+				pw_log_debug(NAME" %p: latency '%s' quantum %u/%u",
+						node, str, quantum_size, context->defaults.clock_rate);
+				pw_log_info("(%s-%u) quantum %u/%u", node->name, node->info.id,
+						quantum_size, context->defaults.clock_rate);
 				node->quantum_size = quantum_size;
 				do_recalc |= node->active;
 			}
@@ -757,7 +838,22 @@ static void check_properties(struct pw_impl_node *node)
 	pw_log_debug(NAME" %p: driver:%d recalc:%d", node, node->driver, do_recalc);
 
 	if (do_recalc)
-		pw_context_recalc_graph(node->context);
+		pw_context_recalc_graph(context, "quantum change");
+}
+
+static const char *str_status(uint32_t status)
+{
+	switch (status) {
+	case PW_NODE_ACTIVATION_NOT_TRIGGERED:
+		return "not-triggered";
+	case PW_NODE_ACTIVATION_TRIGGERED:
+		return "triggered";
+	case PW_NODE_ACTIVATION_AWAKE:
+		return "awake";
+	case PW_NODE_ACTIVATION_FINISHED:
+		return "finished";
+	}
+	return "unknown";
 }
 
 static void dump_states(struct pw_impl_node *driver)
@@ -766,18 +862,19 @@ static void dump_states(struct pw_impl_node *driver)
 
 	spa_list_for_each(t, &driver->rt.target_list, link) {
 		struct pw_node_activation *a = t->activation;
+		struct pw_node_activation_state *state = &a->state[0];
 		if (t->node == NULL)
 			continue;
-		pw_log_warn(NAME" %p (%s): pending:%d/%d s:%"PRIu64" a:%"PRIu64" f:%"PRIu64
-				" waiting:%"PRIu64" process:%"PRIu64" status:%d sync:%d",
-				t->node, t->node->name,
-				a->state[0].pending, a->state[0].required,
+		pw_log_warn("(%s-%u) state:%p pending:%d/%d s:%"PRIu64" a:%"PRIu64" f:%"PRIu64
+				" waiting:%"PRIu64" process:%"PRIu64" status:%s sync:%d",
+				t->node->name, t->node->info.id, state,
+				state->pending, state->required,
 				a->signal_time,
 				a->awake_time,
 				a->finish_time,
 				a->awake_time - a->signal_time,
 				a->finish_time - a->awake_time,
-				a->status, a->pending_sync);
+				str_status(a->status), a->pending_sync);
 	}
 }
 
@@ -797,16 +894,15 @@ static inline int resume_node(struct pw_impl_node *this, int status)
 	pw_log_trace_fp(NAME" %p: trigger peers %"PRIu64, this, nsec);
 
 	spa_list_for_each(t, &this->rt.target_list, link) {
-		struct pw_node_activation_state *state;
+		struct pw_node_activation *a = t->activation;
+		struct pw_node_activation_state *state = &a->state[0];
 
-		state = &t->activation->state[0];
-
-		pw_log_trace_fp(NAME" %p: state %p pending %d/%d", t->node, state,
+		pw_log_trace_fp(NAME" %p: state:%p pending:%d/%d", t->node, state,
                                 state->pending, state->required);
 
 		if (pw_node_activation_state_dec(state, 1)) {
-			t->activation->status = PW_NODE_ACTIVATION_TRIGGERED;
-			t->activation->signal_time = nsec;
+			a->status = PW_NODE_ACTIVATION_TRIGGERED;
+			a->signal_time = nsec;
 			t->signal(t->data);
 		}
 	}
@@ -879,6 +975,9 @@ static inline int process_node(void *data)
 	} else {
 		resume_node(this, status);
 	}
+	if (status & SPA_STATUS_DRAINED) {
+		pw_context_driver_emit_drained(this->context, this);
+	}
 	return 0;
 }
 
@@ -895,8 +994,10 @@ static void node_on_fd_events(struct spa_source *source)
 	if (SPA_LIKELY(source->rmask & SPA_IO_IN)) {
 		uint64_t cmd;
 
-		if (SPA_UNLIKELY(spa_system_eventfd_read(data_system, this->source.fd, &cmd) < 0 || cmd != 1))
-			pw_log_warn(NAME" %p: read %"PRIu64" failed %m", this, cmd);
+		if (SPA_UNLIKELY(spa_system_eventfd_read(data_system, this->source.fd, &cmd) < 0))
+			pw_log_warn(NAME" %p: read failed %m", this);
+		else if (SPA_UNLIKELY(cmd > 1))
+			pw_log_warn(NAME" %p: missed %"PRIu64" wakeups", this, cmd - 1);
 
 		pw_log_trace_fp(NAME" %p: got process", this);
 		this->rt.target.signal(this->rt.target.data);
@@ -946,6 +1047,7 @@ struct pw_impl_node *pw_context_create_node(struct pw_context *context,
 
 	this = &impl->this;
 	this->context = context;
+	this->name = strdup("node");
 
 	if (user_data_size > 0)
                 this->user_data = SPA_MEMBER(impl, sizeof(struct impl), void);
@@ -957,13 +1059,13 @@ struct pw_impl_node *pw_context_create_node(struct pw_context *context,
 		goto error_clean;
 	}
 
-	pw_log_debug(NAME" %p: new", this);
-
 	this->properties = properties;
 
 	if ((res = spa_system_eventfd_create(data_system, SPA_FD_CLOEXEC | SPA_FD_NONBLOCK)) < 0)
 		goto error_clean;
 
+	pw_log_debug(NAME" %p: new fd:%d", this, res);
+
 	this->source.fd = res;
 	this->source.func = node_on_fd_events;
 	this->source.data = this;
@@ -1018,6 +1120,9 @@ struct pw_impl_node *pw_context_create_node(struct pw_context *context,
 	this->rt.activation->sync_timeout = DEFAULT_SYNC_TIMEOUT;
 	this->rt.activation->sync_left = 0;
 
+	this->rt.rate_limit.interval = 2 * SPA_NSEC_PER_SEC;
+	this->rt.rate_limit.burst = 1;
+
 	check_properties(this);
 
 	this->driver_node = this;
@@ -1256,7 +1361,7 @@ static void do_reposition(struct pw_impl_node *driver, struct pw_impl_node *node
 	src = &node->rt.activation->reposition;
 	dst = &a->position.segments[0];
 
-	pw_log_trace(NAME" %p: update position:%lu", node, src->position);
+	pw_log_trace(NAME" %p: update position:%"PRIu64, node, src->position);
 
 	dst->version = src->version;
 	dst->flags = src->flags;
@@ -1288,7 +1393,8 @@ static void update_position(struct pw_impl_node *node, int all_ready)
 
 	if (a->position.state == SPA_IO_POSITION_STATE_STARTING) {
 		if (!all_ready && --a->sync_left == 0) {
-			pw_log_warn(NAME" %p: sync timeout, going to RUNNING", node);
+			pw_log_warn("(%s-%u) sync timeout, going to RUNNING",
+					node->name, node->info.id);
 			pw_context_driver_emit_timeout(node->context, node);
 			dump_states(node);
 			all_ready = true;
@@ -1312,14 +1418,20 @@ static int node_ready(void *data, int status)
 
 	if (SPA_UNLIKELY(node == driver)) {
 		struct pw_node_activation *a = node->rt.activation;
+		struct pw_node_activation_state *state = &a->state[0];
 		int sync_type, all_ready, update_sync, target_sync;
 		uint32_t owner[2], reposition_owner;
 		uint64_t min_timeout = UINT64_MAX;
 
-		if (SPA_UNLIKELY(a->state[0].pending > 0)) {
-			pw_log_warn(NAME" %p: graph not finished: pending %d", node, a->state[0].pending);
+		if (SPA_UNLIKELY(state->pending > 0)) {
 			pw_context_driver_emit_incomplete(node->context, node);
-			dump_states(node);
+			if (ratelimit_test(&node->rt.rate_limit, a->signal_time)) {
+				pw_log_warn("(%s-%u) graph not finished: state:%p quantum:%"PRIu64
+						" pending %d/%d", node->name, node->info.id,
+						state, a->position.clock.duration,
+						state->pending, state->required);
+				dump_states(node);
+			}
 			node->rt.target.signal(node->rt.target.data);
 		}
 
@@ -1401,8 +1513,11 @@ static int node_xrun(void *data, uint64_t trigger, uint64_t delay, struct spa_po
 	a->xrun_delay = delay;
 	a->max_delay = SPA_MAX(a->max_delay, delay);
 
-	pw_log_debug(NAME" %p: XRun! count:%u time:%"PRIu64" delay:%"PRIu64" max:%"PRIu64,
-			this, a->xrun_count, trigger, delay, a->max_delay);
+	if (ratelimit_test(&this->rt.rate_limit, a->signal_time)) {
+		pw_log_error("(%s-%d) XRun! count:%u time:%"PRIu64" delay:%"PRIu64" max:%"PRIu64,
+				this->name, this->info.id, a->xrun_count,
+				trigger, delay, a->max_delay);
+	}
 
 	pw_context_driver_emit_xrun(this->context, this);
 
@@ -1433,22 +1548,9 @@ int pw_impl_node_set_implementation(struct pw_impl_node *node,
 	spa_node_set_callbacks(node->node, &node_callbacks, node);
 	res = spa_node_add_listener(node->node, &node->listener, &node_events, node);
 
-	if (spa_node_set_io(node->node,
-			    SPA_IO_Position,
-			    &node->rt.activation->position,
-			    sizeof(struct spa_io_position)) >= 0) {
-		pw_log_debug(NAME" %p: set position %p", node, &node->rt.activation->position);
-		node->rt.position = &node->rt.activation->position;
-	} else if (node->driver) {
-		pw_log_warn(NAME" %p: can't set position on driver", node);
-	}
-	if (spa_node_set_io(node->node,
-			    SPA_IO_Clock,
-			    &node->rt.activation->position.clock,
-			    sizeof(struct spa_io_clock)) >= 0) {
-		pw_log_debug(NAME" %p: set clock %p", node, &node->rt.activation->position.clock);
-		node->rt.clock = &node->rt.activation->position.clock;
-	}
+	if (node->registered)
+		update_io(node);
+
 	return res;
 }
 
@@ -1481,14 +1583,18 @@ void pw_impl_node_destroy(struct pw_impl_node *node)
 	struct impl *impl = SPA_CONTAINER_OF(node, struct impl, this);
 	struct pw_impl_port *port;
 	struct pw_impl_node *follower;
+	bool active;
 
+	active = node->active;
 	node->active = false;
 
 	pw_log_debug(NAME" %p: destroy", impl);
-	pw_impl_node_emit_destroy(node);
+	pw_log_info("(%s-%u) destroy", node->name, node->info.id);
 
 	suspend_node(node);
 
+	pw_impl_node_emit_destroy(node);
+
 	pw_log_debug(NAME" %p: driver node %p", impl, node->driver_node);
 
 	/* remove ourself as a follower from the driver node */
@@ -1511,12 +1617,6 @@ void pw_impl_node_destroy(struct pw_impl_node *node)
 		spa_node_set_callbacks(node->node, NULL, NULL);
 	}
 
-	pw_log_debug(NAME" %p: unlink ports", node);
-	spa_list_for_each(port, &node->input_ports, link)
-		pw_impl_port_unlink(port);
-	spa_list_for_each(port, &node->output_ports, link)
-		pw_impl_port_unlink(port);
-
 	pw_log_debug(NAME" %p: destroy ports", node);
 	spa_list_consume(port, &node->input_ports, link)
 		pw_impl_port_destroy(port);
@@ -1528,7 +1628,8 @@ void pw_impl_node_destroy(struct pw_impl_node *node)
 		pw_global_destroy(node->global);
 	}
 
-	pw_context_recalc_graph(node->context);
+	if (active)
+		pw_context_recalc_graph(node->context, "active node destroy");
 
 	pw_log_debug(NAME" %p: free", node);
 	pw_impl_node_emit_free(node);
@@ -1762,17 +1863,15 @@ int pw_impl_node_set_state(struct pw_impl_node *node, enum pw_node_state state)
 {
 	int res = 0;
 	struct impl *impl = SPA_CONTAINER_OF(node, struct impl, this);
-	enum pw_node_state old = node->info.state;
+	enum pw_node_state old = impl->pending;
 
 	pw_log_debug(NAME" %p: set state %s -> %s, active %d", node,
 			pw_node_state_as_string(old),
 			pw_node_state_as_string(state),
 			node->active);
 
-	if (old == state)
-		return 0;
-
-	pw_impl_node_emit_state_request(node, state);
+	if (old != state)
+		pw_impl_node_emit_state_request(node, state);
 
 	switch (state) {
 	case PW_NODE_STATE_CREATING:
@@ -1783,7 +1882,7 @@ int pw_impl_node_set_state(struct pw_impl_node *node, enum pw_node_state state)
 		break;
 
 	case PW_NODE_STATE_IDLE:
-		if (node->active && impl->pause_on_idle)
+		if (impl->pause_on_idle)
 			res = pause_node(node);
 		break;
 
@@ -1803,8 +1902,11 @@ int pw_impl_node_set_state(struct pw_impl_node *node, enum pw_node_state state)
 	if (SPA_RESULT_IS_ASYNC(res)) {
 		res = spa_node_sync(node->node, res);
 	}
-	pw_work_queue_add(impl->work,
-			node, res, on_state_complete, SPA_INT_TO_PTR(state));
+	impl->pending = state;
+
+	if (old != state)
+		pw_work_queue_add(impl->work,
+				node, res, on_state_complete, SPA_INT_TO_PTR(state));
 
 	return res;
 }
@@ -1817,17 +1919,12 @@ int pw_impl_node_set_active(struct pw_impl_node *node, bool active)
 	if (old != active) {
 		pw_log_debug(NAME" %p: %s", node, active ? "activate" : "deactivate");
 
-		if (!active)
-			pw_impl_node_set_state(node, PW_NODE_STATE_IDLE);
-
 		node->active = active;
 		pw_impl_node_emit_active_changed(node, active);
 
-		if (active)
-			node_activate(node);
-
 		if (node->registered)
-			pw_context_recalc_graph(node->context);
+			pw_context_recalc_graph(node->context,
+					active ? "node activate" : "node deactivate");
 	}
 	return 0;
 }
