@@ -45,7 +45,7 @@ struct ucred {
 #endif
 
 #ifndef spa_debug
-#define spa_debug pw_log_trace
+#define spa_debug(...) pw_log_trace(__VA_ARGS__)
 #endif
 
 struct defaults {
@@ -59,6 +59,29 @@ struct defaults {
 	unsigned int mem_allow_mlock;
 };
 
+struct ratelimit {
+	uint64_t interval;
+	uint64_t begin;
+	unsigned burst;
+	unsigned n_printed, n_missed;
+};
+
+static inline bool ratelimit_test(struct ratelimit *r, uint64_t now)
+{
+	if (r->begin + r->interval < now) {
+		if (r->n_missed)
+			pw_log_warn("%u events suppressed", r->n_missed);
+		r->begin = now;
+		r->n_printed = 0;
+		r->n_missed = 0;
+	} else if (r->n_printed >= r->burst) {
+		r->n_missed++;
+		return false;
+	}
+	r->n_printed++;
+	return true;
+}
+
 #define MAX_PARAMS	32
 
 #define pw_protocol_emit_destroy(p) spa_hook_list_call(&p->listener_list, struct pw_protocol_events, destroy, 0)
@@ -235,6 +258,7 @@ pw_core_resource_errorf(struct pw_resource *resource, uint32_t id, int seq,
 #define pw_context_driver_emit_xrun(c,n)	pw_context_driver_emit(c, xrun, 0, n)
 #define pw_context_driver_emit_incomplete(c,n)	pw_context_driver_emit(c, incomplete, 0, n)
 #define pw_context_driver_emit_timeout(c,n)	pw_context_driver_emit(c, timeout, 0, n)
+#define pw_context_driver_emit_drained(c,n)	pw_context_driver_emit(c, drained, 0, n)
 
 struct pw_context_driver_events {
 #define PW_VERSION_CONTEXT_DRIVER_EVENTS	0
@@ -248,6 +272,8 @@ struct pw_context_driver_events {
 	void (*incomplete) (void *data, struct pw_impl_node *node);
 	/** The driver got a sync timeout */
 	void (*timeout) (void *data, struct pw_impl_node *node);
+	/** a node drained */
+	void (*drained) (void *data, struct pw_impl_node *node);
 };
 
 #define pw_registry_resource(r,m,v,...) pw_resource_call(r, struct pw_registry_events,m,v,##__VA_ARGS__)
@@ -386,7 +412,7 @@ struct pw_impl_module {
 };
 
 struct pw_node_activation_state {
-	int status;			/**< current status */
+	int status;                     /**< current status, the result of spa_node_process() */
 	int32_t required;		/**< required number of signals */
 	int32_t pending;		/**< number of pending signals */
 };
@@ -522,6 +548,7 @@ struct pw_impl_node {
 					  *  is selected to drive the graph */
 	unsigned int visited:1;		/**< for sorting */
 	unsigned int want_driver:1;	/**< this node wants to be assigned to a driver */
+	unsigned int passive:1;		/**< driver graph only has passive links */
 
 	uint32_t port_user_data_size;	/**< extra size for port user data */
 
@@ -540,20 +567,11 @@ struct pw_impl_node {
 	struct spa_list output_ports;		/**< list of output ports */
 	struct pw_map output_port_map;		/**< map from port_id to port */
 
-	uint32_t n_used_input_links;		/**< number of active input links */
-	uint32_t idle_used_input_links;		/**< number of active input to be idle */
-	uint32_t n_ready_input_links;		/**< number of ready input links */
-
-	uint32_t n_used_output_links;		/**< number of active output links */
-	uint32_t idle_used_output_links;	/**< number of active output to be idle */
-	uint32_t n_ready_output_links;		/**< number of ready output links */
-
 	struct spa_hook_list listener_list;
 
 	struct pw_loop *data_loop;		/**< the data loop for this node */
 
 	uint32_t quantum_size;			/**< desired quantum */
-	uint32_t quantum_current;		/**< current quantum for driver */
 	struct spa_source source;		/**< source to remotely trigger this node */
 	struct pw_memblock *activation;
 	struct {
@@ -570,6 +588,8 @@ struct pw_impl_node {
 		struct pw_node_target target;		/* our target that is signaled by the
 							   driver */
 		struct spa_list driver_link;		/* our link in driver */
+
+		struct ratelimit rate_limit;
 	} rt;
 
         void *user_data;                /**< extra user data */
@@ -618,6 +638,7 @@ struct pw_impl_port_implementation {
 #define pw_impl_port_emit_state_changed(p,o,s,e)	pw_impl_port_emit(p, state_changed, 0, o, s, e)
 #define pw_impl_port_emit_control_added(p,c)		pw_impl_port_emit(p, control_added, 0, c)
 #define pw_impl_port_emit_control_removed(p,c)		pw_impl_port_emit(p, control_removed, 0, c)
+#define pw_impl_port_emit_param_changed(p,i)		pw_impl_port_emit(p, param_changed, 1, i)
 
 #define PW_IMPL_PORT_IS_CONTROL(port)	SPA_FLAG_MASK(port->flags, \
 						PW_IMPL_PORT_FLAG_BUFFERS|PW_IMPL_PORT_FLAG_CONTROL,\
@@ -630,7 +651,7 @@ struct pw_impl_port {
 	struct spa_hook global_listener;
 
 #define PW_IMPL_PORT_FLAG_TO_REMOVE		(1<<0)		/**< if the port should be removed from the
-							  *  implementation when destroyed */
+								  *  implementation when destroyed */
 #define PW_IMPL_PORT_FLAG_BUFFERS		(1<<1)		/**< port has data */
 #define PW_IMPL_PORT_FLAG_CONTROL		(1<<2)		/**< port has control */
 #define PW_IMPL_PORT_FLAG_NO_MIXER		(1<<3)		/**< don't try to add mixer to port */
@@ -659,7 +680,7 @@ struct pw_impl_port {
 	struct spa_callbacks impl;
 
 	struct spa_node *mix;		/**< port buffer mix/split */
-#define PW_IMPL_PORT_MIX_FLAG_MULTI		(1<<0)	/**< multi input or output */
+#define PW_IMPL_PORT_MIX_FLAG_MULTI	(1<<0)	/**< multi input or output */
 #define PW_IMPL_PORT_MIX_FLAG_MIX_ONLY	(1<<1)	/**< only negotiate mix ports */
 #define PW_IMPL_PORT_MIX_FLAG_NEGOTIATE	(1<<2)	/**< negotiate buffers  */
 	uint32_t mix_flags;		/**< flags for the mixing */
@@ -701,19 +722,21 @@ struct pw_control_link {
 
 struct pw_impl_link {
 	struct pw_context *context;		/**< context object */
-	struct spa_list link;		/**< link in context link_list */
-	struct pw_global *global;	/**< global for this link */
+	struct spa_list link;			/**< link in context link_list */
+	struct pw_global *global;		/**< global for this link */
 	struct spa_hook global_listener;
 
-        struct pw_link_info info;		/**< introspectable link info */
+	char *name;
+
+	struct pw_link_info info;		/**< introspectable link info */
 	struct pw_properties *properties;	/**< extra link properties */
 
-	struct spa_io_buffers *io;	/**< link io area */
+	struct spa_io_buffers *io;		/**< link io area */
 
 	struct pw_impl_port *output;		/**< output port */
-	struct spa_list output_link;	/**< link in output port links */
+	struct spa_list output_link;		/**< link in output port links */
 	struct pw_impl_port *input;		/**< input port */
-	struct spa_list input_link;	/**< link in input port links */
+	struct spa_list input_link;		/**< link in input port links */
 
 	struct spa_hook_list listener_list;
 
@@ -722,14 +745,17 @@ struct pw_impl_link {
 
 	struct {
 		struct pw_impl_port_mix out_mix;	/**< port added to the output mixer */
-		struct pw_impl_port_mix in_mix;	/**< port added to the input mixer */
-		struct pw_node_target target;	/**< target to trigger the input node */
+		struct pw_impl_port_mix in_mix;		/**< port added to the input mixer */
+		struct pw_node_target target;		/**< target to trigger the input node */
 	} rt;
 
 	void *user_data;
 
 	unsigned int registered:1;
 	unsigned int feedback:1;
+	unsigned int preparing:1;
+	unsigned int prepared:1;
+	unsigned int passive:1;
 };
 
 #define pw_resource_emit(o,m,v,...) spa_hook_list_call(&o->listener_list, struct pw_resource_events, m, v, ##__VA_ARGS__)
@@ -784,6 +810,7 @@ struct pw_proxy {
 					  *  be removed from server */
 	unsigned int removed:1;		/**< proxy was removed from server */
 	unsigned int destroyed:1;	/**< proxy was destroyed by client */
+	unsigned int in_map:1;		/**< proxy is in core object map */
 
 	struct spa_hook_list listener_list;
 	struct spa_hook_list object_listener_list;
@@ -816,6 +843,7 @@ struct pw_core {
 	int recv_seq;				/**< last received sequence number */
 	int send_seq;				/**< last protocol result code */
 
+	unsigned int removed:1;
 	unsigned int destroyed:1;
 
 	void *user_data;			/**< extra user data */
@@ -959,13 +987,17 @@ pw_context_find_port(struct pw_context *context,
 		  struct spa_pod **format_filters,
 		  char **error);
 
+int pw_context_debug_port_params(struct pw_context *context,
+		struct spa_node *node, enum spa_direction direction,
+		uint32_t port_id, uint32_t id, int err, const char *debug, ...);
+
 const struct pw_export_type *pw_context_find_export_type(struct pw_context *context, const char *type);
 
 int pw_proxy_init(struct pw_proxy *proxy, const char *type, uint32_t version);
 
 void pw_proxy_remove(struct pw_proxy *proxy);
 
-int pw_context_recalc_graph(struct pw_context *context);
+int pw_context_recalc_graph(struct pw_context *context, const char *reason);
 
 void pw_impl_port_update_info(struct pw_impl_port *port, const struct spa_port_info *info);
 
@@ -1062,6 +1094,22 @@ int pw_control_remove_link(struct pw_control_link *link);
 void pw_control_destroy(struct pw_control *control);
 
 void pw_proxy_unref(struct pw_proxy *proxy);
+void pw_proxy_ref(struct pw_proxy *proxy);
+
+#define PW_LOG_OBJECT_POD	(1<<0)
+void pw_log_log_object(enum spa_log_level level, const char *file, int line,
+	   const char *func, uint32_t flags, const void *object);
+
+#define pw_log_object(lev,fl,obj)						\
+({										\
+	if (SPA_UNLIKELY(pw_log_level_enabled (lev)))				\
+		pw_log_log_object(lev,__FILE__,__LINE__,__func__,(fl),(obj));	\
+})
+
+#define pw_log_pod(lev,pod) pw_log_object(lev,PW_LOG_OBJECT_POD,pod)
+#define pw_log_format(lev,pod) pw_log_object(lev,PW_LOG_OBJECT_POD,pod)
+
+bool pw_log_is_default(void);
 
 /** \endcond */
 
