@@ -28,7 +28,9 @@
 #include <signal.h>
 #include <string.h>
 #include <ctype.h>
+#ifndef __FreeBSD__
 #include <alloca.h>
+#endif
 #include <getopt.h>
 
 #include <spa/utils/result.h>
@@ -54,6 +56,7 @@ struct data {
 	struct pw_map vars;
 	unsigned int interactive:1;
 	unsigned int monitoring:1;
+	unsigned int quit:1;
 };
 
 struct global {
@@ -431,6 +434,7 @@ static void on_core_destroy(void *_data)
 
 	pw_map_remove(&data->vars, rd->id);
 	pw_map_for_each(&rd->globals, destroy_global, rd);
+	pw_map_clear(&rd->globals);
 
 	if (data->current == rd)
 		data->current = NULL;
@@ -442,6 +446,12 @@ static const struct pw_proxy_events proxy_core_events = {
 	.destroy = on_core_destroy,
 };
 
+static void remote_data_free(struct remote_data *rd)
+{
+	pw_proxy_destroy((struct pw_proxy*)rd->registry);
+	pw_core_disconnect(rd->core);
+}
+
 static bool do_connect(struct data *data, const char *cmd, char *args, char **error)
 {
 	char *a[1];
@@ -502,7 +512,8 @@ static bool do_disconnect(struct data *data, const char *cmd, char *args, char *
 			goto no_remote;
 
 	}
-	pw_core_disconnect(rd->core);
+	if (rd)
+		remote_data_free(rd);
 
 	if (data->current == NULL) {
 		if (spa_list_is_empty(&data->remotes)) {
@@ -882,7 +893,9 @@ static void client_event_permissions(void *object, uint32_t index,
 			fprintf(stdout, "  default:");
 		else
 			fprintf(stdout, "  %u:", permissions[i].id);
-		fprintf(stdout, " %08x\n", permissions[i].permissions);
+		fprintf(stdout, " %c%c%c\n", permissions[i].permissions & PW_PERM_R ? 'r' : '-',
+					  permissions[i].permissions & PW_PERM_W ? 'w' : '-',
+					  permissions[i].permissions & PW_PERM_X ? 'x' : '-');
 	}
 }
 
@@ -1088,6 +1101,13 @@ static const struct pw_endpoint_stream_events endpoint_stream_events = {
 	.param = event_param
 };
 
+static void
+removed_proxy (void *data)
+{
+	struct proxy_data *pd = data;
+	pw_proxy_destroy(pd->proxy);
+}
+
 static void
 destroy_proxy (void *data)
 {
@@ -1106,6 +1126,7 @@ destroy_proxy (void *data)
 
 static const struct pw_proxy_events proxy_events = {
         PW_VERSION_PROXY_EVENTS,
+        .removed = removed_proxy,
         .destroy = destroy_proxy,
 };
 
@@ -1740,7 +1761,7 @@ children_of(struct remote_data *rd, uint32_t parent_id,
 			if (!count)
 				return 0;
 
-			*children = malloc(sizeof(*children) * count);
+			*children = malloc(sizeof(uint32_t) * count);
 			if (!*children)
 				return -1;
 		}
@@ -1775,10 +1796,6 @@ children_of(struct remote_data *rd, uint32_t parent_id,
 
 		}
 	}
-
-	if (!count)
-		return 0;
-
 	return count;
 }
 
@@ -2706,7 +2723,7 @@ static void do_input(void *data, int fd, uint32_t mask)
 
 	if (mask & SPA_IO_IN) {
 		while (true) {
-			r = read(fd, buf, sizeof(buf));
+			r = read(fd, buf, sizeof(buf)-1);
 			if (r < 0) {
 				if (errno == EAGAIN)
 					continue;
@@ -2740,17 +2757,18 @@ static void do_input(void *data, int fd, uint32_t mask)
 static void do_quit(void *data, int signal_number)
 {
 	struct data *d = data;
+	d->quit = true;
 	pw_main_loop_quit(d->loop);
 }
 
 static void show_help(struct data *data, const char *name)
 {
         fprintf(stdout, "%s [options] [command]\n"
-             "  -h, --help                            Show this help\n"
-             "  -v, --version                         Show version\n"
-             "  -d, --daemon                          Start as daemon (Default false)\n"
-             "  -r, --remote                          Remote daemon name\n\n",
-	     name);
+		"  -h, --help                            Show this help\n"
+		"      --version                         Show version\n"
+		"  -d, --daemon                          Start as daemon (Default false)\n"
+		"  -r, --remote                          Remote daemon name\n\n",
+		name);
 
 	do_help(data, "help", "", NULL);
 }
@@ -2762,23 +2780,24 @@ int main(int argc, char *argv[])
 	char *opt_remote = NULL;
 	char *error;
 	bool daemon = false;
+	struct remote_data *rd;
 	static const struct option long_options[] = {
-		{"help",	0, NULL, 'h'},
-		{"version",	0, NULL, 'v'},
-		{"daemon",	0, NULL, 'd'},
-		{"remote",	1, NULL, 'r'},
-		{NULL,		0, NULL, 0}
+		{ "help",	no_argument,		 NULL, 'h' },
+		{ "version",	no_argument,		 NULL, 'V' },
+		{ "daemon",	no_argument,		 NULL, 'd' },
+		{ "remote",	required_argument,	 NULL, 'r' },
+		{ NULL,	0, NULL, 0}
 	};
 	int c, i;
 
 	pw_init(&argc, &argv);
 
-	while ((c = getopt_long(argc, argv, "hvdr:", long_options, NULL)) != -1) {
+	while ((c = getopt_long(argc, argv, "hVdr:", long_options, NULL)) != -1) {
 		switch (c) {
 		case 'h':
 			show_help(&data, argv[0]);
 			return 0;
-		case 'v':
+		case 'V':
 			fprintf(stdout, "%s\n"
 				"Compiled with libpipewire %s\n"
 				"Linked with libpipewire %s\n",
@@ -2793,6 +2812,7 @@ int main(int argc, char *argv[])
 			opt_remote = optarg;
 			break;
 		default:
+			show_help(&data, argv[0]);
 			return -1;
 		}
 	}
@@ -2813,7 +2833,10 @@ int main(int argc, char *argv[])
 
 	pw_context_load_module(data.context, "libpipewire-module-link-factory", NULL, NULL);
 
-	do_connect(&data, "connect", opt_remote, &error);
+	if (!do_connect(&data, "connect", opt_remote, &error)) {
+		fprintf(stderr, "Error: \"%s\"\n", error);
+		return -1;
+	}
 
 	if (optind == argc) {
 		data.interactive = true;
@@ -2839,14 +2862,18 @@ int main(int argc, char *argv[])
 			fprintf(stdout, "Error: \"%s\"\n", error);
 			free(error);
 		}
-		if (data.current) {
+		if (!data.quit && data.current) {
 			data.current->prompt_pending = pw_core_sync(data.current->core, 0, 0);
 			pw_main_loop_run(data.loop);
 		}
 	}
+	spa_list_consume(rd, &data.remotes, link)
+		remote_data_free(rd);
 
 	pw_context_destroy(data.context);
 	pw_main_loop_destroy(data.loop);
+	pw_map_clear(&data.vars);
+	pw_deinit();
 
 	return 0;
 }
